# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
# flake8: noqa

from __future__ import absolute_import

from base64 import b64decode

multipart_samples = []
formencoded_samples = []

# 1k block of random bytes
FILE_UPLOAD_DATA = b64decode(b'''
QJHX77oVa5GoHaIthmjSe8i8r6AhNIpIYx9Yw46gDqifTyd14aE/hWGoTCMR0hjFcE0GAQ7KE/S7
wnpoaAMManjCR4wD39lzhUsUPLy1N/ANwAEo0fk93y8ggEhbN8lEi0zl5rIecC1ZsEA8WG4dw3gh
61FRZdZ/fG8sojnfAjWM9UnSdhkd6idvF8qwKGBRpFPFUT4gRNu+di4i9r6+XQ7IiXbc2KXiXyy0
7AsrcOsOOiincLkw/IvDVYUgZmUiTqs0XKHfsFexpznIVIKubYGGlGpIyBdq04VmtPoQvj9weqZh
di5F9jlGBJ479vHNS+6xwbnxqpSNkAkzsBOQ6a68Ae9in1R0t+UIlUlRKG6jsUelkbJlq3+qnzqt
RFBuOneh/vdlFEhuXQSxZ4a7oPl+wuv94fE6AsWyiV4urZR16CaioAbeozrd5MOqBkKk7Q2TyWX/
hHA05mlv6g4/34lh5N4UWH8YTLEjgL9ykMJ6zodmGujmvmYwIiVQuhIcjsZ5n0m0zYBlDP4BDspV
bQsTxPa+0ptiQJ+qTStqPWCDaNFO6Fwdia7jxs7asPAuCwIcgUL2ztmCfafvlniMSwC8lmjs77Zt
gvgeCWr6IoASWOso1oEFsCgk6HMEU+XW5bLD5h5cva0/on/K1wwpFF+FQlQEqvd2cAEFZrZ8Dk4E
zKK+tdU3qK5maooKTIwpj8llk3ENpRAZXJUYRpUzvdb3m4kDg+sHCbZw5t8bpC5Oph89AHytc+Op
zY6+/UUHiUuIuf0hflVqQgb9uangMAkHkMyjw6o7dFPPanfu2j0DouqGF2QKRczdcs9itbWh65/O
PX1KbztEqAAxTqZixNI9qkG8A9rW24lDX6QFSws/0Wjh/UUvixAq9u/Dobbqn+hUgsQl2xqoMtci
sRJajasOPIyIOgwpLP5ul2DMIqZUDmAii3OcOQL4y4b0vgUHk26iNp9omPpzenJuAECoyIE8KnEf
SK1Cv9mQOU6Gbl3gyLsMp9fdsespzLFv8M+VrlaLCDTgL6OL2DW9SwOH4FWKH8qAjJfTsWnhqpbS
PBRsHlujXX1lAZC5eKCA4ekug2+aW7g+Mpse2Rt1XP4w8UZ231bxCYmYsKrzsB5f/0Pfp/vAjYHm
GYtj6RBhRoTym2Sm/A1VOOOBYSRPZXeqdx2WoIrGtSlkEZnH2ySM5z1Se6v9y4uI0TN5Oa5q691G
UegOAO9Yx4oa1L+7BXs+lkIqz94bVCLNI6gwLC+QOhz6iC96LyagKZtcYyagLeiQGhh91oJXWCCm
sz4EbqNbiEtfTa6LhAs+mkLSOfrAsyWLr/reTu4uHwJXsGzgn17Uhe/+VxV74/yEuwTLTbHHgA==
''')

# Mozilla/5.0 (X11; U; FreeBSD i386; en-US; rv:1.9.0.11) Gecko/2009061410 Firefox/3.0.11
multipart_samples.append({'content_length': '1638', 'data': b'-----------------------------1282332249122276511499183892\r\nContent-Disposition: form-data; name="empty-text-input"\r\n\r\n\r\n-----------------------------1282332249122276511499183892\r\nContent-Disposition: form-data; name="text-input-ascii"\r\n\r\nabcdef\r\n-----------------------------1282332249122276511499183892\r\nContent-Disposition: form-data; name="text-input-unicode"\r\n\r\n\xce\xb1\xce\xb2\xce\xb3\xce\xb4\r\n-----------------------------1282332249122276511499183892\r\nContent-Disposition: form-data; name="file-upload"; filename="test.data"\r\nContent-Type: application/octet-stream\r\n\r\n@\x91\xd7\xef\xba\x15k\x91\xa8\x1d\xa2-\x86h\xd2{\xc8\xbc\xaf\xa0!4\x8aHc\x1fX\xc3\x8e\xa0\x0e\xa8\x9fO\'u\xe1\xa1?\x85a\xa8L#\x11\xd2\x18\xc5pM\x06\x01\x0e\xca\x13\xf4\xbb\xc2zhh\x03\x0cjx\xc2G\x8c\x03\xdf\xd9s\x85K\x14<\xbc\xb57\xf0\r\xc0\x01(\xd1\xf9=\xdf/ \x80H[7\xc9D\x8bL\xe5\xe6\xb2\x1ep-Y\xb0@<Xn\x1d\xc3x!\xebQQe\xd6\x7f|o,\xa29\xdf\x025\x8c\xf5I\xd2v\x19\x1d\xea\'o\x17\xca\xb0(`Q\xa4S\xc5Q> D\xdb\xbev."\xf6\xbe\xbe]\x0e\xc8\x89v\xdc\xd8\xa5\xe2_,\xb4\xec\x0b+p\xeb\x0e:(\xa7p\xb90\xfc\x8b\xc3U\x85 fe"N\xab4\\\xa1\xdf\xb0W\xb1\xa79\xc8T\x82\xaem\x81\x86\x94jH\xc8\x17j\xd3\x85f\xb4\xfa\x10\xbe?pz\xa6av.E\xf69F\x04\x9e;\xf6\xf1\xcdK\xee\xb1\xc1\xb9\xf1\xaa\x94\x8d\x90\t3\xb0\x13\x90\xe9\xae\xbc\x01\xefb\x9fTt\xb7\xe5\x08\x95IQ(n\xa3\xb1G\xa5\x91\xb2e\xab\x7f\xaa\x9f:\xadDPn:w\xa1\xfe\xf7e\x14Hn]\x04\xb1g\x86\xbb\xa0\xf9~\xc2\xeb\xfd\xe1\xf1:\x02\xc5\xb2\x89^.\xad\x94u\xe8&\xa2\xa0\x06\xde\xa3:\xdd\xe4\xc3\xaa\x06B\xa4\xed\r\x93\xc9e\xff\x84p4\xe6io\xea\x0e?\xdf\x89a\xe4\xde\x14X\x7f\x18L\xb1#\x80\xbfr\x90\xc2z\xce\x87f\x1a\xe8\xe6\xbef0"%P\xba\x12\x1c\x8e\xc6y\x9fI\xb4\xcd\x80e\x0c\xfe\x01\x0e\xcaUm\x0b\x13\xc4\xf6\xbe\xd2\x9bb@\x9f\xaaM+j=`\x83h\xd1N\xe8\\\x1d\x89\xae\xe3\xc6\xce\xda\xb0\xf0.\x0b\x02\x1c\x81B\xf6\xce\xd9\x82}\xa7\xef\x96x\x8cK\x00\xbc\x96h\xec\xef\xb6m\x82\xf8\x1e\tj\xfa"\x80\x12X\xeb(\xd6\x81\x05\xb0($\xe8s\x04S\xe5\xd6\xe5\xb2\xc3\xe6\x1e\\\xbd\xad?\xa2\x7f\xca\xd7\x0c)\x14_\x85BT\x04\xaa\xf7vp\x01\x05f\xb6|\x0eN\x04\xcc\xa2\xbe\xb5\xd57\xa8\xaefj\x8a\nL\x8c)\x8f\xc9e\x93q\r\xa5\x10\x19\\\x95\x18F\x953\xbd\xd6\xf7\x9b\x89\x03\x83\xeb\x07\t\xb6p\xe6\xdf\x1b\xa4.N\xa6\x1f=\x00|\xads\xe3\xa9\xcd\x8e\xbe\xfdE\x07\x89K\x88\xb9\xfd!~UjB\x06\xfd\xb9\xa9\xe00\t\x07\x90\xcc\xa3\xc3\xaa;tS\xcfjw\xee\xda=\x03\xa2\xea\x86\x17d\nE\xcc\xddr\xcfb\xb5\xb5\xa1\xeb\x9f\xce=}Jo;D\xa8\x001N\xa6b\xc4\xd2=\xaaA\xbc\x03\xda\xd6\xdb\x89C_\xa4\x05K\x0b?\xd1h\xe1\xfdE/\x8b\x10*\xf6\xef\xc3\xa1\xb6\xea\x9f\xe8T\x82\xc4%\xdb\x1a\xa82\xd7"\xb1\x12Z\x8d\xab\x0e<\x8c\x88:\x0c),\xfen\x97`\xcc"\xa6T\x0e`"\x8bs\x9c9\x02\xf8\xcb\x86\xf4\xbe\x05\x07\x93n\xa26\x9fh\x98\xfaszrn\x00@\xa8\xc8\x81<*q\x1fH\xadB\xbf\xd9\x909N\x86n]\xe0\xc8\xbb\x0c\xa7\xd7\xdd\xb1\xeb)\xcc\xb1o\xf0\xcf\x95\xaeV\x8b\x084\xe0/\xa3\x8b\xd85\xbdK\x03\x87\xe0U\x8a\x1f\xca\x80\x8c\x97\xd3\xb1i\xe1\xaa\x96\xd2<\x14l\x1e[\xa3]}e\x01\x90\xb9x\xa0\x80\xe1\xe9.\x83o\x9a[\xb8>2\x9b\x1e\xd9\x1bu\\\xfe0\xf1Fv\xdfV\xf1\t\x89\x98\xb0\xaa\xf3\xb0\x1e_\xffC\xdf\xa7\xfb\xc0\x8d\x81\xe6\x19\x8bc\xe9\x10aF\x84\xf2\x9bd\xa6\xfc\rU8\xe3\x81a$Oew\xaaw\x1d\x96\xa0\x8a\xc6\xb5)d\x11\x99\xc7\xdb$\x8c\xe7=R{\xab\xfd\xcb\x8b\x88\xd13y9\xaej\xeb\xddFQ\xe8\x0e\x00\xefX\xc7\x8a\x1a\xd4\xbf\xbb\x05{>\x96B*\xcf\xde\x1bT"\xcd#\xa80,/\x90:\x1c\xfa\x88/z/&\xa0)\x9b\\c&\xa0-\xe8\x90\x1a\x18}\xd6\x82WX \xa6\xb3>\x04n\xa3[\x88K_M\xae\x8b\x84\x0b>\x9aB\xd29\xfa\xc0\xb3%\x8b\xaf\xfa\xdeN\xee.\x1f\x02W\xb0l\xe0\x9f^\xd4\x85\xef\xfeW\x15{\xe3\xfc\x84\xbb\x04\xcbM\xb1\xc7\x80\r\n-----------------------------1282332249122276511499183892--\r\n', 'content_type': 'multipart/form-data; boundary=---------------------------1282332249122276511499183892'})

# Mozilla/5.0 (X11; U; FreeBSD i386; en-US; rv:1.9.0.11) Gecko/2009061410 Firefox/3.0.11
formencoded_samples.append({'content_length': '85', 'data': b'empty-text-input=&text-input-ascii=abcdef&text-input-unicode=%CE%B1%CE%B2%CE%B3%CE%B4', 'content_type': 'application/x-www-form-urlencoded'})

# Mozilla/4.0 (compatible; MSIE 7.0; Windows NT 5.1; InfoPath.2)
multipart_samples.append({'content_length': '1616', 'data': b'-----------------------------7d91772e200be\r\nContent-Disposition: form-data; name="empty-text-input"\r\n\r\n\r\n-----------------------------7d91772e200be\r\nContent-Disposition: form-data; name="text-input-ascii"\r\n\r\nabcdef\r\n-----------------------------7d91772e200be\r\nContent-Disposition: form-data; name="text-input-unicode"\r\n\r\n\xce\xb1\xce\xb2\xce\xb3\xce\xb4\r\n-----------------------------7d91772e200be\r\nContent-Disposition: form-data; name="file-upload"; filename="C:\\Documents and Settings\\Administrator\\My Documents\\test.data"\r\nContent-Type: application/octet-stream\r\n\r\n@\x91\xd7\xef\xba\x15k\x91\xa8\x1d\xa2-\x86h\xd2{\xc8\xbc\xaf\xa0!4\x8aHc\x1fX\xc3\x8e\xa0\x0e\xa8\x9fO\'u\xe1\xa1?\x85a\xa8L#\x11\xd2\x18\xc5pM\x06\x01\x0e\xca\x13\xf4\xbb\xc2zhh\x03\x0cjx\xc2G\x8c\x03\xdf\xd9s\x85K\x14<\xbc\xb57\xf0\r\xc0\x01(\xd1\xf9=\xdf/ \x80H[7\xc9D\x8bL\xe5\xe6\xb2\x1ep-Y\xb0@<Xn\x1d\xc3x!\xebQQe\xd6\x7f|o,\xa29\xdf\x025\x8c\xf5I\xd2v\x19\x1d\xea\'o\x17\xca\xb0(`Q\xa4S\xc5Q> D\xdb\xbev."\xf6\xbe\xbe]\x0e\xc8\x89v\xdc\xd8\xa5\xe2_,\xb4\xec\x0b+p\xeb\x0e:(\xa7p\xb90\xfc\x8b\xc3U\x85 fe"N\xab4\\\xa1\xdf\xb0W\xb1\xa79\xc8T\x82\xaem\x81\x86\x94jH\xc8\x17j\xd3\x85f\xb4\xfa\x10\xbe?pz\xa6av.E\xf69F\x04\x9e;\xf6\xf1\xcdK\xee\xb1\xc1\xb9\xf1\xaa\x94\x8d\x90\t3\xb0\x13\x90\xe9\xae\xbc\x01\xefb\x9fTt\xb7\xe5\x08\x95IQ(n\xa3\xb1G\xa5\x91\xb2e\xab\x7f\xaa\x9f:\xadDPn:w\xa1\xfe\xf7e\x14Hn]\x04\xb1g\x86\xbb\xa0\xf9~\xc2\xeb\xfd\xe1\xf1:\x02\xc5\xb2\x89^.\xad\x94u\xe8&\xa2\xa0\x06\xde\xa3:\xdd\xe4\xc3\xaa\x06B\xa4\xed\r\x93\xc9e\xff\x84p4\xe6io\xea\x0e?\xdf\x89a\xe4\xde\x14X\x7f\x18L\xb1#\x80\xbfr\x90\xc2z\xce\x87f\x1a\xe8\xe6\xbef0"%P\xba\x12\x1c\x8e\xc6y\x9fI\xb4\xcd\x80e\x0c\xfe\x01\x0e\xcaUm\x0b\x13\xc4\xf6\xbe\xd2\x9bb@\x9f\xaaM+j=`\x83h\xd1N\xe8\\\x1d\x89\xae\xe3\xc6\xce\xda\xb0\xf0.\x0b\x02\x1c\x81B\xf6\xce\xd9\x82}\xa7\xef\x96x\x8cK\x00\xbc\x96h\xec\xef\xb6m\x82\xf8\x1e\tj\xfa"\x80\x12X\xeb(\xd6\x81\x05\xb0($\xe8s\x04S\xe5\xd6\xe5\xb2\xc3\xe6\x1e\\\xbd\xad?\xa2\x7f\xca\xd7\x0c)\x14_\x85BT\x04\xaa\xf7vp\x01\x05f\xb6|\x0eN\x04\xcc\xa2\xbe\xb5\xd57\xa8\xaefj\x8a\nL\x8c)\x8f\xc9e\x93q\r\xa5\x10\x19\\\x95\x18F\x953\xbd\xd6\xf7\x9b\x89\x03\x83\xeb\x07\t\xb6p\xe6\xdf\x1b\xa4.N\xa6\x1f=\x00|\xads\xe3\xa9\xcd\x8e\xbe\xfdE\x07\x89K\x88\xb9\xfd!~UjB\x06\xfd\xb9\xa9\xe00\t\x07\x90\xcc\xa3\xc3\xaa;tS\xcfjw\xee\xda=\x03\xa2\xea\x86\x17d\nE\xcc\xddr\xcfb\xb5\xb5\xa1\xeb\x9f\xce=}Jo;D\xa8\x001N\xa6b\xc4\xd2=\xaaA\xbc\x03\xda\xd6\xdb\x89C_\xa4\x05K\x0b?\xd1h\xe1\xfdE/\x8b\x10*\xf6\xef\xc3\xa1\xb6\xea\x9f\xe8T\x82\xc4%\xdb\x1a\xa82\xd7"\xb1\x12Z\x8d\xab\x0e<\x8c\x88:\x0c),\xfen\x97`\xcc"\xa6T\x0e`"\x8bs\x9c9\x02\xf8\xcb\x86\xf4\xbe\x05\x07\x93n\xa26\x9fh\x98\xfaszrn\x00@\xa8\xc8\x81<*q\x1fH\xadB\xbf\xd9\x909N\x86n]\xe0\xc8\xbb\x0c\xa7\xd7\xdd\xb1\xeb)\xcc\xb1o\xf0\xcf\x95\xaeV\x8b\x084\xe0/\xa3\x8b\xd85\xbdK\x03\x87\xe0U\x8a\x1f\xca\x80\x8c\x97\xd3\xb1i\xe1\xaa\x96\xd2<\x14l\x1e[\xa3]}e\x01\x90\xb9x\xa0\x80\xe1\xe9.\x83o\x9a[\xb8>2\x9b\x1e\xd9\x1bu\\\xfe0\xf1Fv\xdfV\xf1\t\x89\x98\xb0\xaa\xf3\xb0\x1e_\xffC\xdf\xa7\xfb\xc0\x8d\x81\xe6\x19\x8bc\xe9\x10aF\x84\xf2\x9bd\xa6\xfc\rU8\xe3\x81a$Oew\xaaw\x1d\x96\xa0\x8a\xc6\xb5)d\x11\x99\xc7\xdb$\x8c\xe7=R{\xab\xfd\xcb\x8b\x88\xd13y9\xaej\xeb\xddFQ\xe8\x0e\x00\xefX\xc7\x8a\x1a\xd4\xbf\xbb\x05{>\x96B*\xcf\xde\x1bT"\xcd#\xa80,/\x90:\x1c\xfa\x88/z/&\xa0)\x9b\\c&\xa0-\xe8\x90\x1a\x18}\xd6\x82WX \xa6\xb3>\x04n\xa3[\x88K_M\xae\x8b\x84\x0b>\x9aB\xd29\xfa\xc0\xb3%\x8b\xaf\xfa\xdeN\xee.\x1f\x02W\xb0l\xe0\x9f^\xd4\x85\xef\xfeW\x15{\xe3\xfc\x84\xbb\x04\xcbM\xb1\xc7\x80\r\n-----------------------------7d91772e200be--\r\n', 'content_type': 'multipart/form-data; boundary=---------------------------7d91772e200be'})

# Mozilla/4.0 (compatible; MSIE 7.0; Windows NT 5.1; InfoPath.2)
formencoded_samples.append({'content_length': '85', 'data': b'empty-text-input=&text-input-ascii=abcdef&text-input-unicode=%CE%B1%CE%B2%CE%B3%CE%B4', 'content_type': 'application/x-www-form-urlencoded'})



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
from __future__ import absolute_import

import pytest

from fresco import FrescoApp, GET, Response
from fresco.requestcontext import RequestContext


class TestRequestContext(object):

    def test_instantiation(self):
        """
        Can we cleanly instantiate a RequestContext object?
        """
        RequestContext()

    def test_app_populates_request_object(self):

        def view():
            from fresco.core import context
            assert context.request is not None
            return Response([''])

        app = FrescoApp()
        app.route('/', GET, view)

        with app.requestcontext('/'):
            app.view()

    def test_context_returns_correct_request_for_each_app(self):

        from time import sleep
        from threading import Thread, current_thread

        threadcount = 3
        itercount = 200
        calls = []

        def view():
            from fresco.core import context
            request_id = id(context.request)

            def generate_response():
                for i in range(itercount):
                    assert id(context.request) == request_id
                    calls.append(current_thread().ident)
                    sleep(0.0001)
                    yield str(request_id).encode('ascii')
            return Response(list(generate_response()))

        app = FrescoApp()
        app.route('/', GET, view)

        def do_request():
            with app.requestcontext('/'):
                app.view()

        threads = [Thread(target=do_request) for i in range(threadcount)]

        for t in threads:
            t.start()
        for t in threads:
            t.join()

        assert len(calls) == threadcount * itercount

        # Check that threaded requests were genuinely interleaved
        for i1, i2 in zip(calls[:itercount - 1], calls[1:itercount]):
            if i1 != i2:
                break
        else:
            raise AssertionError("Output does not appear interleaved")

    def test_context_does_not_inherit_from_parent(self):
        c = RequestContext()
        c.push(foo=1)
        c.push(bar=2)
        with pytest.raises(AttributeError):
            c.foo

    def test_child_context_overrides_parent(self):
        c = RequestContext()
        c.push(foo=1)
        c.push(foo=2)
        assert c.foo == 2

    def test_pop_context_removes_keys(self):
        c = RequestContext()
        c.push(foo=1)
        c.push(foo=2)
        assert c.foo == 2
        c.pop()
        assert c.foo == 1



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
from __future__ import absolute_import

from mock import Mock, call
import pytest

from fresco import Route, GET, POST, FrescoApp, Response
from fresco import routeargs
from fresco.exceptions import BadRequest


class TestRouteArg(object):

    def test_routekwarg_configured(self):

        A = routeargs.RouteArg()
        route = Route('/', GET, lambda r: None, x=A)
        assert A.route is route
        assert A.name is 'x'

    def test_routekwarg_value_passed(self):

        view = Mock(return_value=Response())
        routekwarg = Mock(spec=routeargs.RouteArg)
        routekwarg.return_value = 'xyzzy'

        app = FrescoApp()
        app.route('/', GET, view, x=routekwarg)
        with app.requestcontext('/') as c:
            app.view()
            assert routekwarg.call_args_list == [call(c.request)]
            assert view.call_args_list == [call(x='xyzzy')]

    def test_queryarg_value_passed(self):
        view = Mock(return_value=Response())
        app = FrescoApp()
        app.route('/', GET, view, x=routeargs.QueryArg())
        with app.requestcontext('/?x=foo'):
            app.view()
            assert view.call_args_list == [call(x='foo')]

    def test_formarg_value_passed(self):
        view = Mock(return_value=Response())
        app = FrescoApp()
        app.route('/', GET, view, x=routeargs.FormArg())
        with app.requestcontext('/?x=foo'):
            app.view()
            assert view.call_args_list == [call(x='foo')]

    def test_sessionarg_value_passed(self):
        view = Mock(return_value=Response())
        app = FrescoApp()
        app.route('/', GET, view, x=routeargs.SessionArg())
        with app.requestcontext('/?x=foo') as c:
            c.request.environ[c.request.SESSION_ENV_KEY] = {'x': 'foo'}
            app.view()
            assert view.call_args_list == [call(x='foo')]

    def test_cookiearg_value_passed(self):
        view = Mock(return_value=Response())
        app = FrescoApp()
        app.route('/', GET, view, x=routeargs.CookieArg())
        with app.requestcontext('/', HTTP_COOKIE='x=foo'):
            app.view()
            assert view.call_args_list == [call(x='foo')]

    def test_cookiearg_listvalue_passed(self):
        view = Mock(return_value=Response())
        app = FrescoApp()
        app.route('/', GET, view, x=routeargs.CookieArg([str]))
        with app.requestcontext('/', HTTP_COOKIE='x=foo;x=bar'):
            app.view()
            assert view.call_args_list == [call(x=['foo', 'bar'])]

    def test_requestarg_value_converted(self):
        view = Mock(return_value=Response())
        app = FrescoApp()
        app.route('/', GET, view, x=routeargs.FormArg(float))
        with app.requestcontext('/?x=0'):
            app.view()
            assert view.call_args_list == [call(x=0.0)]

    def test_requestarg_default_value(self):
        view = Mock(return_value=Response())
        app = FrescoApp()
        app.route('/', GET, view, x=routeargs.FormArg(default='d'))
        with app.requestcontext('/'):
            app.view()
            assert view.call_args_list == [call(x='d')]

    def test_it_doesnt_convert_default_values(self):
        view = Mock(return_value=Response())
        app = FrescoApp()
        app.route('/', GET, view, x=routeargs.FormArg(int, default=None))
        with app.requestcontext('/'):
            app.view()
            assert view.call_args_list == [call(x=None)]

    def test_access_to_missing_requestarg_returns_badrequest(self):
        def view(x):
            x == 0
        app = FrescoApp()
        app.route('/', GET, view, x=routeargs.FormArg())
        with app.requestcontext('/'):
            assert 'Bad Request' in app.view().status

    def test_missing_requestarg_with_conversion_returns_badrequest(self):
        def view(x):
            x == 0
        app = FrescoApp()
        app.route('/', GET, view, x=routeargs.FormArg(int))
        with app.requestcontext('/'):
            assert 'Bad Request' in app.view().status

    def test_routeargs_work_as_positional_arguments(self):
        view = Mock(return_value=Response())
        app = FrescoApp()
        app.route('/', GET, view, args=[routeargs.FormArg(key='x')])
        with app.requestcontext('/?x=foo'):
            app.view()
            assert view.call_args_list == [call('foo')]

    def test_routearg_classes_are_auto_instantiated(self):
        view = Mock(return_value=Response())
        app = FrescoApp()
        app.route('/', GET, view, x=routeargs.FormArg)
        with app.requestcontext('/?x=foo'):
            app.view()
            assert view.call_args_list == [call(x='foo')]


class TestRequestObject(object):

    def test_it_passes_the_request(self):
        view = Mock(return_value=Response())
        app = FrescoApp()
        app.route('/', GET, view, x=routeargs.RequestObject())
        with app.requestcontext('/') as c:
            app.view()
            x = view.call_args[1]['x']
            assert x is c.request


class Test_routearg(object):

    def test_it_calls_func(object):

        view = Mock(return_value=Response())
        argfunc = Mock()
        app = FrescoApp()
        app.route('/', GET, view, x=routeargs.routearg(argfunc))
        with app.requestcontext('/'):
            app.view()
            assert argfunc.call_count == 1, argfunc.call_count
            assert view.call_args[1]['x'] is argfunc()

    def test_it_passes_additional_args(object):

        view = Mock(return_value=Response())
        argfunc = Mock()
        app = FrescoApp()
        app.route('/', GET, view, x=routeargs.routearg(argfunc, 'x', y=42))
        with app.requestcontext('/') as c:
            app.view()
            assert argfunc.call_args == ((c.request, 'x'), {'y': 42})
            assert view.call_args[1]['x'] is argfunc()


class TestFormData(object):

    def test_it_passes_the_form_dict(self):
        view = Mock(return_value=Response())
        app = FrescoApp()
        app.route('/', GET, view, x=routeargs.FormData())
        with app.requestcontext('/') as c:
            app.view()
            x = view.call_args[1]['x']
            assert x is c.request.form


class TestJSONPayload(object):

    def test_it_passes_json_data(self):
        view = Mock(return_value=Response())
        app = FrescoApp()
        app.route('/', POST, view, x=routeargs.JSONPayload())
        with app.requestcontext_post('/', data=b'{"foo":"bar"}',
                                     CONTENT_TYPE='application/json'):
            app.view()
            x = view.call_args[1]['x']
            assert x == {'foo': 'bar'}

    def test_it_raises_bad_request_on_invalid_payload(self):

        view = Mock(return_value=Response())
        app = FrescoApp()
        app.route('/', POST, view, x=routeargs.JSONPayload())
        with app.requestcontext_post('/', data=b'invalid json!',
                                     CONTENT_TYPE='application/json'):
            with pytest.raises(BadRequest):
                app.view()

    def test_it_returns_a_default_value(self):
        view = Mock(return_value=Response())
        app = FrescoApp()
        app.route('/', POST, view,
                  x=routeargs.JSONPayload(default='default value'))
        with app.requestcontext_post('/'):
            app.view()
        x = view.call_args[1]['x']
        assert x == 'default value'



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
from __future__ import absolute_import

import sys
import inspect
from operator import attrgetter
from functools import wraps, partial
from itertools import cycle
import warnings

from fresco import GET, POST, PUT, Response
from fresco.exceptions import BadRequest

_marker = object()


def onerror(exceptions, handler):
    """\
    Return a decorator that can replace or update the return value of the
    function if an exception is raised
    """

    try:
        if isinstance(exceptions, Exception):
            exceptions = (exceptions,)
    except TypeError:
        pass

    def decorator(func):
        @wraps(func)
        def decorated(*fargs, **fkwargs):
            try:
                return func(*fargs, **fkwargs)
            except exceptions:
                exc_info = sys.exc_info()
                return handler(exc_info, *fargs, **fkwargs)
        return decorated
    return decorator


def json_response(data=_marker,
                  indent=None,
                  separators=(',', ':'),
                  content_type='application/json',
                  **kwargs):
    """
    JSON encode the function's result and return a response object with the
    content-type ``application/json``.

    May also be used as a regular function to directly create a JSON encoded
    response, however it's preferrable to use
    :meth:`~fresco.response.Response.json` for this.

    :param data: The data to json encode.
                 If left unspecified, ``json_response``
                 will act as a function decorator.
    :param indent: The indent level. Defaults to ``None`` (no pretty printing)
    :param separators: Defaults to ``(',', ':')`` for the most compact JSON
                       representation
    :param kwargs: Other keyword arguments are passed to ``json.dumps``. These
                   may be used to change encoding paramters, for example
                   overriding the default ``JSONEncoder`` class.


    """
    # Called as a decorator or decorator factory
    if data is _marker or callable(data):
        def json_response_decorator(func):
            @wraps(func)
            def json_response_decorated(*fa, **fkw):
                return Response.json(
                    func(*fa, **fkw), indent, separators, **kwargs)
            return json_response_decorated

        # Called as a decorator factory (``@json_response()``)
        if data is _marker:
            return json_response_decorator

        # Called as a decorator (ie no parens: ``@json_response``)
        return json_response_decorator(data)

    # Called as a regular function (or via ``Route(...).filter``)
    else:
        if isinstance(data, Response):
            return data

        return Response.json(data,
                            indent=indent,
                            separators=separators,
                            content_type=content_type,
                            **kwargs)


def extract_requestargs(requesttype, **argspec):
    """
    Decorator mapping request querystring/post data to function arguments.

    """
    warnings.warn("extract_requestargs is deprecated. "
                  "Please see fresco.routeargs for alternatives",
                  DeprecationWarning, stacklevel=2)
    request_attr = {
        GET: attrgetter('query'),
        POST: attrgetter('form'),
        PUT: attrgetter('form'),
    }[requesttype]

    strict_checking = argspec.pop('_strict', False)
    exception = argspec.pop('_raise_on_error', BadRequest)

    def decorator(func):
        """
        Decorate function ``func``.
        """

        f_args, f_varargs, f_varkw, f_defaults = inspect.getargspec(func)

        # Produce a mapping {argname: default}
        if f_defaults is None:
            f_defaults = []
        defaults = dict(zip(f_args[-len(f_defaults):], f_defaults))

        def decorated(*args, **kwargs):
            """
            Call ``func`` with arguments extracted from ``request``.
            """

            from fresco.core import context
            request = context.request
            given_arguments = dict(
                zip(f_args[:len(args)], args)
            )
            given_arguments.update(kwargs)
            newargs = given_arguments.copy()

            for name, type_fn in argspec.items():
                try:
                    try:
                        value = given_arguments[name]
                    except KeyError:
                        if isinstance(type_fn, list):
                            value = request_attr(request).getlist(name)
                        else:
                            value = request_attr(request)[name]

                    try:
                        if isinstance(type_fn, list):
                            value = [cast(v)
                                     for cast, v in zip(cycle(type_fn), value)]
                        else:
                            value = type_fn(value)
                    except ValueError as e:
                        if name in defaults and not strict_checking:
                            value = defaults[name]
                        else:
                            raise exception(
                                "Could not convert parameter %r to "
                                " requested type (%s)" % (name, e.args[0])
                            )

                except KeyError:
                    try:
                        value = defaults[name]
                    except KeyError:
                        raise exception(name)
                newargs[name] = value

            return func(**newargs)
        return wraps(func)(decorated)
    return decorator

extract_getargs = partial(extract_requestargs, GET)
extract_postargs = partial(extract_requestargs, POST)
extract_getargs.__doc__ = """\
        Extract function arguments from request parameters. See extract_getargs
        for details.
        """
extract_getargs.__doc__ = """\
    Extract function arguments from request parameters

    Synopsis::

        >>> from flea import Agent
        >>> from fresco import FrescoApp, GET
        >>>
        >>> app = FrescoApp()
        >>> @app.route('/', GET)
        ... @extract_getargs(GET, id=int)
        ... def my_app(id):
        ...     return Response(['Got id %d' % (id,)])
        ...
        >>> print Agent(app).get('/?id=2')
        200 OK\r
        Content-Type: text/html; charset=UTF-8\r
        \r
        Got id 2

    If specified arguments are not present in the request (and no default value
    is given in the function signature), or a ValueError is thrown during type
    conversion an appropriate error will be raised::

        >>> Agent(app.wsgi_app).get('/') #doctest: +ELLIPSIS
        Traceback (most recent call last):
            ...
        BadStatusError: GET '/' returned HTTP status '400 Bad Request'
        >>> Agent(app.wsgi_app).get('/?id=ratatouille') #doctest: +ELLIPSIS
        Traceback (most recent call last):
            ...
        BadStatusError: GET '/' returned HTTP status '400 Bad Request'

    You can change the error raised, for example if you want a 404 page to be
    returned::

        >>> from fresco.exceptions import NotFound
        >>>
        >>> app = FrescoApp()
        >>> @app.route('/', GET)
        ... @extract_getargs(actions=[unicode])
        ... def my_app(id):
        ...     return Response(['Got id %d' % (id,)])
        ...
        >>> Agent(app).get('/')
        Traceback (most recent call last):
            ...
        BadStatusError: GET '/' returned HTTP status '404 Not Found'

    A default argument value in the handler function will protect against
    this::

        >>> app = FrescoApp()
        >>> @app.route('/', GET)
        ... @extract_getargs(actions=[unicode])
        ... def my_app(id=1):
        ...     return Response(['Got id %d' % (id,)])
        ...
        >>> print Agent(app).get('/')
        200 OK\r
        Content-Type: text/html; charset=UTF-8\r
        \r
        Got id 1

    Sometimes it is necessary to map multiple request values to a single
    argument, for example in a form where two or more input fields have the
    same name. To do this, put the type-casting function into a list when
    calling ``with_request_args``::

        >>> app = FrescoApp()
        >>> @app.route('/', GET)
        ... @extract_getargs(actions=[unicode])
        ... def my_app(actions):
        ...     return Response([
        ...         ', '.join(actions)
        ...     ])
        ...
        >>> print Agent(app).get(
        ...     '/?actions=up;actions=up;actions=and+away%21')
        200 OK\r
        Content-Type: text/html; charset=UTF-8\r
        \r
        up, up, and away!

"""



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
from __future__ import absolute_import

from fresco import FrescoApp, context, GET, Response
from fresco.middleware import XForwarded


class TestXForwarded(object):

    def get_app(self, *args, **kwargs):
        app = FrescoApp()

        @app.route('/', GET)
        def view():
            request = context.request
            return Response([request.url, request.environ['REMOTE_ADDR']])
        app.add_middleware(XForwarded, *args, **kwargs)
        return app

    def test_forwards_x_forwarded_for(self):

        app = self.get_app()
        with app.requestcontext('/',
                                REMOTE_ADDR='127.0.0.1',
                                HTTP_X_FORWARDED_FOR='1.2.3.4'):
            url, addr = app.view().content_iterator
            assert addr == b'1.2.3.4'

    def test_forwards_x_forwarded_host(self):

        app = self.get_app()
        with app.requestcontext('/',
                                REMOTE_ADDR='127.0.0.1',
                                HTTP_X_FORWARDED_HOST='frontendserver'):
            url, addr = app.view().content_iterator
        assert url == b'http://frontendserver/'

    def test_forwards_x_forwarded_ssl(self):

        app = self.get_app()
        with app.requestcontext('/',
                                REMOTE_ADDR='127.0.0.1',
                                HTTP_X_FORWARDED_SSL='on',
                                HTTP_X_FORWARDED_HOST='localhost'):
            url, addr = app.view().content_iterator
            assert url == b'https://localhost/'

    def test_reports_first_trusted_ip(self):
        app = self.get_app(trusted=['127.0.0.1', '3.3.3.3'])
        with app.requestcontext('/',
                                REMOTE_ADDR='127.0.0.1',
                                HTTP_X_FORWARDED_FOR='1.1.1.1, '
                                                     '2.2.2.2, 3.3.3.3'):
            url, addr = app.view().content_iterator
            assert addr == b'2.2.2.2'

    def test_reports_first_listed_ip_as_remote_addr_when_untrusted(self):
        app = self.get_app()
        with app.requestcontext('/',
                                REMOTE_ADDR='127.0.0.1',
                                HTTP_X_FORWARDED_FOR='1.1.1.1, 2.2.2.2'):
            url, addr = app.view().content_iterator
            assert addr == b'1.1.1.1'



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
from __future__ import absolute_import

import re
import sys
from copy import copy
from collections import defaultdict, namedtuple, Iterable
from weakref import WeakKeyDictionary

from fresco.compat import string_types
from fresco.exceptions import ResponseException
from fresco.response import Response
from fresco.requestcontext import context
from fresco.routeargs import RouteArg
from fresco.util.common import fq_path
from fresco.util.urls import join_path

RFC2518_METHODS = \
    PROPFIND, PROPPATCH, MKCOL, COPY, MOVE, LOCK, UNLOCK = \
    'PROPFIND', 'PROPPATCH', 'MKCOL', 'COPY', 'MOVE', 'LOCK', 'UNLOCK'

RFC2616_METHODS = \
    GET, HEAD, POST, PUT, DELETE, OPTIONS, TRACE, CONNECT = \
    'GET', 'HEAD', 'POST', 'PUT', 'DELETE', 'OPTIONS', 'TRACE', 'CONNECT'

RFC3253_METHODS = \
    VERSION_CONTROL, REPORT, CHECKOUT, CHECKIN, UNCHECKOUT, MKWORKSPACE,\
    UPDATE, LABEL, MERGE, BASELINE_CONTROL, MKACTIVITY = \
    'VERSION-CONTROL', 'REPORT', 'CHECKOUT', 'CHECKIN', 'UNCHECKOUT', \
    'MKWORKSPACE', 'UPDATE', 'LABEL', 'MERGE', 'BASELINE-CONTROL', \
    'MKACTIVITY',

RFC3648_METHODS = ORDERPATCH, = ('ORDERPATCH',)

RFC3744_METHODS = ACL, = ('ACL',)

RFC5323_METHODS = SEARCH, = ('SEARCH',)

RFC5789_METHODS = PATCH, = ('PATCH',)

ALL_METHODS = HTTP_METHODS = set(RFC2518_METHODS + RFC2616_METHODS +
                                 RFC3253_METHODS + RFC3648_METHODS +
                                 RFC5323_METHODS + RFC5789_METHODS)

__all__ = ('ALL_METHODS', 'Pattern', 'Route', 'DelegateRoute',
           'RouteCollection', 'routefor')

__all__ += tuple(method.replace('-', '_') for method in ALL_METHODS)


#: Encapsulate a pattern match on a path
PathMatch = namedtuple('PathMatch', ['path_matched', 'path_remaining',
                                     'args', 'kwargs'])


class RouteTraversal(namedtuple('RouteTraversal',
                                'route args kwargs collections_traversed')):
    """
    Encapsulate a route traversal.

    Each ``RouteTraversal`` object contains:

    - ``route`` the final route traversed, allowing access to the view
        associated with the path.
    - ``args`` - positional args to be passed to the view. This is a
        combination of args extracted from the path and any added when
        constructing the route.
    - ``kwargs`` - keyword args to be passed to the view. This is a
        combination of kwargs extracted from the path and any added when
        constructing the route.
    -  ``collections_traversed`` - a list of ``TraversedCollection``
        objects, containing information on which RouteCollections were
        encountered during traversal, the route selected at each stage
        and any args/kwargs associated with that phase of the traversal.
    """

    def replace(self, viewspec, traversal_kwargs=None, **kwargs):
        """
        Return a new ``RouteTraversal`` with the traversal_kwargs or other
        fields replaced for the route identified by ``viewspec``.

        Example:

            >>> from fresco import FrescoApp
            >>> app = FrescoApp()
            >>> @route('/<lang:str>/index.html', GET, homepage, name='home')
            ... def homepage(lang):
            ...     return Response({'fr': 'Bonjour!', 'en': 'Hello!'}[lang])
            ...
            >>> traversal = next(app.get_routes('/fr/index.html'))
            >>> en_traversal = traversal.replace('home', {'lang': 'en'})
            >>> en_traversal.build_path()
            '/en/index.html'

        """
        if traversal_kwargs is not None:
            kwargs['traversal_kwargs'] = traversal_kwargs
        if isinstance(viewspec, string_types):
            viewspecs = viewspec.split(':')
        else:
            viewspecs = [viewspec]
        collections_traversed = iter(self.collections_traversed)
        for item in viewspecs:
            while True:
                ct = next(collections_traversed, None)
                if ct is None:
                    raise RouteNotFound(viewspec)
                try:
                    route = ct.collection.routefor(item)
                    break
                except RouteNotFound:
                    continue

        collections_traversed = [c._replace(**kwargs)
                                 if (c.route is route) else c
                                 for c in self.collections_traversed]
        return self._replace(collections_traversed=collections_traversed)

    def build_path(self):
        """
        Reconstruct the path from a route traversal
        """
        paths = []
        for c in self.collections_traversed:
            path, _, _ = c.route.path(*c.traversal_args,
                                      **c.traversal_kwargs)
            paths.append(path)
        return ''.join(paths)

#: An item of RouteTraversal.collections_traversed
TraversedCollection = namedtuple('TraversedCollection',
                                 'collection path route args kwargs '
                                 'traversal_args traversal_kwargs')


_marker = object()


class RouteNotReady(RuntimeError):
    """
    getview was called on a route mapping to an instance method, but the
    class has yet been instantiated
    """


class URLGenerationError(Exception):
    """\
    Was not possible to generate the requested URL.
    """


class RouteNotFound(Exception):
    """\
    The named route does not exist in the RouteCollection.
    """


class Pattern(object):
    """\
    Patterns are matchable against URL paths using their ``match`` method. If a
    path matches, this should return a tuple of ``(positional_arguments,
    keyword_arguments)`` extracted from the URL path. Otherwise this method
    should return ``None``.

    Pattern objects may also be able to take a tuple of
    ``(positional_arguments, keyword_arguments)`` and return a corresponding
    URL path.
    """

    def match(self, path):
        """
        Should return a tuple of ``(positional_arguments, keyword_arguments)``
        if the pattern matches the given URL path, or None if it does not
        match.
        """
        raise NotImplementedError

    def pathfor(self, *args, **kwargs):
        """
        The inverse of ``match``, this should return a URL path
        for the given positional and keyword arguments, along with any unused
        arguments.

        :return: a tuple of ``(path, remaining_args, remaining_kwargs)``
        """
        raise NotImplementedError()

    def path_argument_info(self):
        """
        Return information about the arguments required for ``pathfor``
        """
        raise NotImplementedError()

    def add_prefix(self, prefix):
        """
        Return a copy of the pattern with the given string prepended
        """
        raise NotImplementedError()


class Converter(object):
    """\
    Responsible for converting arguments to and from URL components.

    A ``Converter`` class should provide two instance methods:

    - ``to_string``: convert from a python object to a string
    - ``from_string``: convert from URL-encoded bytestring to the target
                        python type.

    It must also define the regular expression pattern that is used to extract
    the string from the URL.
    """
    pattern = '[^/]+'

    def __init__(self, pattern=None):
        """
        Initialize a ``Converter`` instance.
        """
        if pattern is not None:
            self.pattern = pattern

    def to_string(self, ob):
        """
        Convert arbitrary argument ``ob`` to a string representation
        """
        return str(ob)

    def from_string(self, s):
        """
        Convert string argument ``s`` to the target object representation,
        whatever that may be.
        """
        return s


class IntConverter(Converter):
    """\
    Match any integer value and convert to an ``int`` value.
    """

    pattern = r'[+-]?\d+'

    def from_string(self, s):
        """
        Return ``s`` converted to an ``int`` value.
        """
        return int(s)


class StrConverter(Converter):
    """\
    Match any string, not including a forward slash, and return a ``str``
    value
    """

    pattern = r'[^/]+'

    def to_string(self, s):
        """
        Return ``s`` converted to an ``str`` object.
        """
        return s

    def from_string(self, s):
        """
        Return ``s`` converted to a (unicode) string type
        """
        return s


class AnyConverter(StrConverter):
    """
    Match any one of the given string options. Example::

        "/<lang:any('fr', 'en', 'de')>"
    """
    pattern = None

    def __init__(self, *args):
        super(AnyConverter, self).__init__(None)
        if len(args) == 0:
            raise ValueError("Must supply at least one argument to any()")
        self.pattern = '|'.join(re.escape(arg) for arg in args)


class PathConverter(StrConverter):
    """\
    Match any string, possibly including forward slashes, and return a
    ``str`` object.
    """
    pattern = r'.+'


class MatchAllURLsPattern(Pattern):
    """\
    A pattern matcher that matches all URLs starting with the given prefix. No
    arguments are parsed from the URL.
    """

    def __init__(self, path):
        self.path = path

    def match(self, path):
        if path.startswith(self.path):
            return PathMatch(self.path, path[len(self.path):], (), {})
        return None

    def pathfor(self, *args, **kwargs):
        assert not args and not kwargs, \
            "MatchAllURLsPattern does not support URL arguments"

        return self.path, (), {}

    def path_argument_info(self):
        return (), {}

    def add_prefix(self, prefix):
        return self.__class__(join_path(prefix, self.path))

    def __str__(self):
        return '%s*' % (self.path,)


class ExtensiblePattern(Pattern):
    """\
    An extensible URL pattern matcher.

    Synopsis::

        >>> from pprint import pprint
        >>> p = ExtensiblePattern(r"/<:str>/<year:int>/<title:str>")
        >>> pprint(p.match('/archive/1999/blah')) # doctest: +ELLIPSIS
        PathMatch(...)

    Patterns are split on slashes into components. A component can either be a
    literal part of the path, or a pattern component in the form::

        <identifier>:<converter>

    ``identifer`` can be any python name, which will be used as the name of a
    keyword argument to the matched function. If omitted, the argument will be
    passed as a positional arg.

    ``converter`` must be the name of a pre-registered converter. Converters
    must support ``to_string`` and ``from_string`` methods and are used to
    convert between URL segments and python objects.

    By default, the following converters are configured:

    - ``int`` - converts to an integer
    - ``path`` - any path (ie can include forward slashes)
    - ``str`` - any string (not including forward slashes)
    - ``unicode`` - alias for ``str``
    - ``any`` - a string matching a list of alternatives

    Some examples::

        >>> p = ExtensiblePattern(r"/images/<:path>")
        >>> p.match('/images/thumbnails/02.jpg') # doctest: +ELLIPSIS
        PathMatch(..., args=('thumbnails/02.jpg',), kwargs={})

        >>> p = ExtensiblePattern("/<page:any('about', 'help')>.html")
        >>> p.match('/about.html') # doctest: +ELLIPSIS
        PathMatch(..., args=(), kwargs={'page': 'about'})

        >>> p = ExtensiblePattern("/entries/<id:int>")
        >>> p.match('/entries/23')  # doctest: +ELLIPSIS
        PathMatch(..., args=(), kwargs={'id': 23})

    Others can be added by calling ``ExtensiblePattern.register_converter``
    """

    preset_patterns = (
        ('int', IntConverter),
        ('str', StrConverter),
        ('unicode', StrConverter),
        ('path', PathConverter),
        ('any', AnyConverter),
    )
    pattern_parser = re.compile("""
        <
            (?P<name>\w[\w\d]*)?
            :
            (?P<converter>\w[\w\d]*)
            (?:
                \(
                         (?P<args>.*?)
                \)
            )?
        >
    """, re.X)

    class Segment(object):
        """
        Represent a single segment of a URL, storing information about hte
        ``source``, ``regex`` used to pattern match the segment, ``name`` for
        named parameters and the ``converter`` used to map the value to a URL
        parameter if applicable
        """

        def __init__(self, source, regex, name, converter):
            self.source = source
            self.regex = regex
            self.name = name
            self.converter = converter

    def __init__(self, pattern, match_entire_path=True):
        """
        Initialize a new ``ExtensiblePattern`` object with pattern ``pattern``

        :param pattern: The pattern string, eg ``'/<id:int>/show'``
        :param match_entire_path: Boolean. If ``True``, the entire path will be
                                  required to match, otherwise a prefix match
                                  will suffice.
        """
        super(ExtensiblePattern, self).__init__()

        self.preset_patterns = dict(self.preset_patterns)
        self.pattern = pattern
        self.match_entire_path = match_entire_path

        self.segments = list(self._make_segments(pattern))
        self.args = [item
                     for item in self.segments
                     if item.converter is not None]

        regex = ''.join(segment.regex for segment in self.segments)
        if self.match_entire_path:
            regex += '$'
        else:
            regex += '(?=/|$)'

        self.regex = re.compile(regex)

    def path_argument_info(self):
        positional = tuple(a.converter for a in self.args if a.name is None)
        keyword = dict((a.name, a.converter)
                       for a in self.args if a.name is not None)
        return (positional, keyword)

    def _make_segments(self, s):
        r"""
        Generate successive Segment objects from the given string.

        Each segment object represents a part of the pattern to be matched, and
        comprises ``source``, ``regex``, ``name`` (if a named parameter) and
        ``converter`` (if a parameter)
        """

        for item in split_iter(self.pattern_parser, self.pattern):
            if isinstance(item, string_types):
                yield self.Segment(item, re.escape(item), None, None)
                continue
            groups = item.groupdict()
            name, converter, args = (groups['name'], groups['converter'],
                                     groups['args'])
            converter = self.preset_patterns[converter]
            if args:
                args, kwargs = self.parseargs(args)
                converter = converter(*args, **kwargs)
            else:
                converter = converter()
            yield self.Segment(item.group(0),
                               '(%s)' % converter.pattern, name, converter)

    def parseargs(self, argstr):
        """
        Return a tuple of ``(args, kwargs)`` parsed out of a string in the
        format ``arg1, arg2, param=arg3``.

        Synopsis::

            >>> ep =  ExtensiblePattern('')
            >>> ep.parseargs("1, 2, 'buckle my shoe'")
            ((1, 2, 'buckle my shoe'), {})
            >>> ep.parseargs("3, four='knock on the door'")
            ((3,), {'four': 'knock on the door'})

        """
        return eval('(lambda *args, **kwargs: (args, kwargs))(%s)' % argstr)

    def match(self, path):
        """
        Test ``path`` and return a tuple of parsed ``(args, kwargs)``, or
        ``None`` if there was no match.
        """
        mo = self.regex.match(path)
        if not mo:
            return None
        groups = mo.groups()
        assert len(groups) == len(self.args), (
            "Number of regex groups does not match expected count. "
            "Perhaps you have used capturing parentheses somewhere? "
            "The pattern matched was %r." % self.regex.pattern
        )

        try:
            groups = [(segment.name, segment.converter.from_string(value))
                      for value, segment in zip(groups, self.args)]
        except ValueError:
            return None

        args = tuple(value for name, value in groups if not name)
        kwargs = dict((name, value) for name, value in groups if name)
        return PathMatch(mo.group(0), path[len(mo.group(0)):], args, kwargs)

    def pathfor(self, *args, **kwargs):
        """
        Example usage::

            >>> p = ExtensiblePattern("/view/<name:str>/<revision:int>")
            >>> p.pathfor(name='important_document.pdf', revision=299)
            ('/view/important_document.pdf/299', [], {})

            >>> p = ExtensiblePattern("/view/<:str>/<:int>")
            >>> p.pathfor('important_document.pdf', 299)
            ('/view/important_document.pdf/299', [], {})
        """

        args = list(args)
        kwargs = kwargs
        result = []
        for seg in self.segments:
            if not seg.converter:
                result.append(seg.source)

            elif seg.name:
                try:
                    value = kwargs.pop(seg.name)
                except IndexError:
                    raise URLGenerationError(
                        "Argument %r not specified for url %r" % (
                            seg.name, self.pattern
                        )
                    )
                result.append(seg.converter.to_string(value))

            else:
                try:
                    value = args.pop(0)
                except IndexError:
                    raise URLGenerationError(
                        "Not enough positional arguments for url %r" % (
                            self.pattern,
                        )
                    )
                result.append(seg.converter.to_string(value))

        return ''.join(result), args, kwargs

    def add_prefix(self, prefix):
        return self.__class__(join_path(prefix, self.pattern),
                              self.match_entire_path)

    @classmethod
    def register_converter(cls, name, converter):
        r"""
        Register a preset pattern for later use in URL patterns.

        Example usage::

            >>> from datetime import date
            >>> from time import strptime
            >>> class DateConverter(Converter):
            ...     pattern = r'\d{8}'
            ...     def from_string(self, s):
            ...         return date(*strptime(s, '%d%m%Y')[:3])
            ...
            >>> ExtensiblePattern.register_converter('date', DateConverter)
            >>> ExtensiblePattern('/<:date>')\
            ...      .match('/01011970')  # doctest:+ELLIPSIS
            PathMatch(..., args=(datetime.date(1970, 1, 1),), kwargs={})
        """
        cls.preset_patterns += ((name, converter),)

    def __repr__(self):
        return '<%s %r>' % (self.__class__, self.pattern)

    def __str__(self):
        return '%s' % (self.pattern,)


class Route(object):
    """\
    Represent a URL routing pattern
    """

    #: The default class to use for URL pattern matching
    pattern_class = ExtensiblePattern

    def __init__(self, pattern, methods=None, view=None, kwargs=None,
                 args=None, name=None, predicate=None, decorators=None,
                 filters=None, **_kwargs):

        """
        :param pattern:     A string that can be compiled into a path pattern
        :param methods:     The list of HTTP methods the view is bound to
                            ('GET', 'POST', etc)
        :param view:        The view function.
        :param kwargs:      A dictionary of default keyword arguments to pass
                            to the view callable
        :param args:        Positional arguments to pass to the view callable
        :param name:        A name that can later be used to retrieve the route
                            for URL generation
        :param predicate:   A callable that is used to decide whether to match
                            this pattern. The callable must take a ``Request``
                            object as its only parameter and return a boolean.
        :param decorators:  Decorator functions to apply to the view callable
                            before invoking it
        :param filters:     Filter functions to apply to the view's return
                            value before returning the final response object
        :param **_kwargs:   Keyword arguments matching HTTP method names
                            (GET, POST etc) can used to specify views
                            associated with those methods.
                            Other keyword aruments are passed through to the
                            view callable.

        Naming routes
        -------------

        Naming routes allows you to reference routes throughout your
        application, eg when generating URLs with
        :function:`~fresco.core.urlfor`.

        If you don't specify a ``name`` argument, the route will available by
        its fully qualified function name (eg
        ``'package.module.view_function'``), or by passing the function object
        itself to ``urlfor``

        Views may often be assigned to multiple routes, for example::

            >>> def display_image(size):
            ...     return Response(['image data...'])
            ...
            >>> from fresco import FrescoApp
            >>> app = FrescoApp()
            >>> app.route('/image', display_image, size=(1024, 768))
            >>> app.route('/thumbnail', display_image, size=(75, 75))

        If you generate a URL for the view function ``display_image``, the
        the first declared route will always win, in this case::

            >>> app.urlfor(display_image)
            'http://localhost/image'

        To generate URLs for the thumbnail route in this example, you must
        explicity assign names::

            >>> app.route('/image', display_image, name='image')
            >>> app.route('/thumbnail', display_image, name='image-thumbnail')

            >>> app.urlfor('image')
            'http://localhost/image'
            >>> app.urlfor('image-thumbnail')
            'http://localhost/thumbnail'

        """
        method_view_map = {}
        if methods:
            if isinstance(methods, string_types):
                methods = [methods]
            else:
                # Catch the common error of not providing a valid method
                if not isinstance(methods, Iterable):
                    raise TypeError("HTTP methods must be specified as a "
                                    "string or iterable")
            for m in methods:
                if m not in ALL_METHODS:
                    raise ValueError(
                        "{!r} is not a valid HTTP method".format(m))
            method_view_map.update((m, view) for m in methods)

        for method in ALL_METHODS:
            if method in _kwargs:
                method_view_map[method] = _kwargs.pop(method)

        if not isinstance(pattern, Pattern):
            pattern = self.pattern_class(pattern)

        if name and ':' in name:
            # Colons are reserved to act as separators
            raise ValueError("Route names cannot contain ':'")

        self.name = name
        self.predicate = predicate
        self.decorators = decorators or []
        self.filters = filters or []

        #: Default values to use for path generation
        self.routed_args_default = {}

        self.pattern = pattern
        self.methods = set(method_view_map)
        self.instance = None

        # Cached references to view functions
        self._cached_views = {}

        # Cached references to decorated view function. We use weakrefs in case
        # a process_view hook substitutes the view function used as a key
        # for a dynamically generated function.
        self._cached_dviews = WeakKeyDictionary()

        p_args, p_kwargs = pattern.path_argument_info()
        for k in p_kwargs:
            default = _kwargs.pop(k + '_default', _marker)
            if default is not _marker:
                self.routed_args_default[k] = default

        self.view_args = tuple(args or _kwargs.pop('view_args', tuple()))
        self.view_kwargs = dict(kwargs or _kwargs.pop('view_kwargs', {}),
                                **_kwargs)

        for arg in self.view_args:
            if isinstance(arg, RouteArg):
                arg.configure(self, None)

        for argname, arg in self.view_kwargs.items():
            # Allow 'Route(... x=FormArg)' shortcut
            if isinstance(arg, type) and issubclass(arg, RouteArg):
                arg = self.view_kwargs[argname] = arg()
            if isinstance(arg, RouteArg):
                arg.configure(self, argname)

        #: A mapping of HTTP methods to view specifiers
        self.viewspecs = method_view_map

    def __repr__(self):
        view_methods_map = defaultdict(set)
        for method, viewspec in self.viewspecs.items():
            view_methods_map[viewspec].add(method)

        s = []
        for viewspec, methods in view_methods_map.items():
            if methods == ALL_METHODS:
                methods = '*'
            else:
                methods = ' '.join(self.methods)
            s.append('%s %s => %r' % (methods, str(self.pattern), fq_path(viewspec)))

        return '<%s %s>' % (self.__class__.__name__,
                            '\n      '.join(s))

    def __getstate__(self):
        state = self.__dict__.copy()
        state['_cached_views'] = {}
        state['_cached_dviews'] = WeakKeyDictionary()
        return state

    def match(self, path, method):
        if method and method not in self.methods:
            return None
        return self.pattern.match(path)

    def getview(self, method):
        """\
        Return the raw view callable.
        """
        try:
            return self._cached_views[method]
        except KeyError:
            pass

        uview = self.viewspecs[method]
        if isinstance(uview, string_types):
            if self.instance is None:
                raise RouteNotReady()
            uview = getattr(self.instance, uview)

        self._cached_views[method] = uview
        return uview

    def getdecoratedview(self, view):
        """\
        Return the view callable decorated with any decorators defined in the
        route
        """
        try:
            return self._cached_dviews[view]
        except KeyError:
            pass

        dview = view
        for d in (self.decorators or []):
            dview = d(dview)
        if self.filters:
            dview = self._apply_filters(dview)
        self._cached_dviews[view] = dview
        return dview

    def _apply_filters(self, view):
        """
        Decorate ``view`` so that the result is passed through any available
        filters.
        """
        def filtered_view(*args, **kwargs):
            result = view(*args, **kwargs)
            for f in self.filters:
                result = f(result)
            return result
        return filtered_view

    def bindto(self, instance):
        """\
        Return copy of the route bound to a given instance.

        Use this when traversing view classes.
        """
        ob = copy(self)
        ob._setinstance(instance)
        return ob

    def _setinstance(self, instance):
        self.instance = instance

    def add_prefix(self, path):
        """
        Return a copy of the Route object with the given path prepended to the
        routing pattern.
        """
        newroute = object.__new__(self.__class__)
        newroute.__dict__ = dict(self.__dict__,
                                 pattern=self.pattern.add_prefix(path))
        return newroute

    def path(self, *args, **kwargs):
        """\
        Build the path corresponding to this route and return a tuple of:

            ``(<path>, <remaining args>, <remaining kwargs>)``

        ``remaining args`` and ``remaining kwargs`` are any values from
        ``*args`` and ``**kwargs`` not consumed during path construction.

        See also :meth:`~fresco.routing.Pattern.pathfor`.
        """
        request = kwargs.pop('request', None)
        for k in self.routed_args_default:
            if k not in kwargs:
                v = self.routed_args_default[k]
                if callable(v):
                    v = v(request)
                kwargs[k] = v
        return self.pattern.pathfor(*args, **kwargs)

    def route_keys(self):
        """\
        Generate keys by which the route should be indexed
        """
        if self.name:
            yield self.name

        for method, viewspec in self.viewspecs.items():
            yield viewspec
            try:
                view = self.getview(method)
            except RouteNotReady:
                continue
            yield view

            # Also return the underlying function object for python2 unbound
            # methods
            func_ob = getattr(view, '__func__', None)
            if func_ob is not None:
                yield func_ob

    def wrap(self, *decorators):
        """
        Wrap the view function in a decorator. Can be chained for a fluid api::

            Route('/user/<id:int>', GET, edituser)\
                    .decorate(require_ssl)
                    .decorate(logged_in)
        """
        self.decorators.extend(decorators)
        return self

    #: ``decorate`` is an alias for :meth:`~fresco.routing.Route.wrap`
    decorate = wrap

    def filter(self, *funcs):
        """
        Filter the output of the view function through other functions::

            Route('/user/<id:int>', GET, edituser)\
                    .decorate(require_ssl)
                    .filter(render('user.tmpl'))
        """
        self.filters.extend(funcs)
        return self


def split_iter(pattern, string):
    """
    Generate alternate strings and match objects for all occurances of
    ``pattern`` in ``string``.
    """
    matcher = pattern.finditer(string)
    match = None
    pos = 0
    for match in matcher:
        yield string[pos:match.start()]
        yield match
        pos = match.end()
    yield string[pos:]


def routefor(viewspec, _app=None):
    """\
    Convenience wrapper around :meth:`~fresco.routing.RouteCollection.urlfor`.
    """
    return context.app.routefor(viewspec)


class RouteCollection(object):
    """\
    A collection of :class:`~fresco.routing.Route` objects, RouteCollections:

    - Contain methods to configure routes, including the ability to
      delegate URLs to other RouteCollections
    - Can map from a request to a view
    """

    route_class = Route

    def __init__(self, routes=None, route_class=None):
        self.__routes__ = []
        self.__routed_views__ = {}
        self.route_class = route_class or self.route_class
        if routes is not None:
            for item in routes:
                if isinstance(item, Route):
                    self.add_route(item)
                elif hasattr(item, '__routes__'):
                    for r in item.__routes__:
                        if r.instance is None:
                            r = r.bindto(item)
                        self.add_route(r)
                elif isinstance(item, Iterable):
                    for r in item:
                        self.add_route(r)
                else:
                    raise TypeError(item)

    def __add__(self, other):
        result = copy(self)
        if isinstance(other, Route):
            result.add_route(other)
        else:
            for item in other:
                result.add_route(item)
        return result

    def __radd__(self, other):
        if isinstance(other, Route):
            return RouteCollection([other]) + self
        elif isinstance(other, Iterable):
            return RouteCollection(other) + self
        raise TypeError("Cannot add %r to %r" % (other, self))

    def __iter__(self):
        return iter(self.__routes__)

    def add_route(self, route):
        self.__routes__.append(route)

    def add_prefix(self, prefix):
        """
        Return a copy of the RouteCollection with the given path prepended to
        all routes.
        """
        ob = copy(self)
        ob.__routes__ = [route.add_prefix(prefix) for route in self]
        return ob

    def pathfor(self, viewspec, *args, **kwargs):
        """\
        Return the path component of the url for the given view name/function
        spec.

        :param viewspec: a view name, a reference in the form
                         ``'package.module.viewfunction'``, or the view
                         callable itself.
        """
        request = kwargs.pop('request', None)
        if isinstance(viewspec, string_types) and ':' in viewspec:
            viewspec, remainder = viewspec.split(':', 1)
            delegated_route = self.routefor(viewspec)

            p1, remainder_args, remainder_kwargs = \
                delegated_route.path(*args, **kwargs)

            factory_args = args[:-len(remainder_args)]
            factory_kwargs = dict((k, v)
                                  for k, v in kwargs.items()
                                  if k not in remainder_kwargs)
            for k in delegated_route.routed_args_default:
                if k not in factory_kwargs:
                    v = delegated_route.routed_args_default[k]
                    if callable(v):
                        v = v(request)
                    factory_kwargs[k] = v
            rc = delegated_route.routecollectionfactory(*factory_args,
                                                        **factory_kwargs)
            p2 = rc.pathfor(remainder, request=request, *args, **kwargs)
            return p1 + p2

        return self.routefor(viewspec).path(request=request,
                                            *args, **kwargs)[0]

    def routefor(self, viewspec):
        """
        Return the :class:`~fresco.routing.Route` instance associated with
        ``viewspec``.

        Views may have multiple routes bound, in this case the first bound
        route will always take precedence.

        This method does not resolve delegated routes.

        :param viewspec: a view callable or a string in the form
                         'package.module.viewfunction'
        """
        try:
            return self.__routed_views__[viewspec]
        except KeyError:
            pass

        for route in self.__routes__:
            for k in route.route_keys():
                self.__routed_views__.setdefault(k, route)

        try:
            return self.__routed_views__[viewspec]
        except KeyError:
            pass

        if not isinstance(viewspec, string_types):
            raise RouteNotFound(viewspec)

        modname, symbols = viewspec, []
        while True:
            try:
                modname, sym = modname.rsplit('.', 1)
            except ValueError:
                raise RouteNotFound(viewspec)

            symbols.append(sym)
            module = sys.modules.get(modname, None)
            if module:
                ob = module
                for s in reversed(symbols):
                    ob = getattr(ob, s)
                try:
                    route = self.__routed_views__[ob]
                except KeyError:
                    # An unbound class method? (python 2 only)
                    try:
                        route = self.__routed_views__[ob.__func__]
                    except (AttributeError, KeyError):
                        raise RouteNotFound(viewspec)

                self.__routed_views__[viewspec] = route
                return route

    def get_routes(self, path, method, request=None):
        """
        Generate routes matching the given path and method::

            for rt in routecollection.get_routes('/foo/bar', GET):
                print("Route is", rt.route)
                print("Arguments extracted from the path:", rt.args, rt.kwargs)
                print("RouteCollections are:", rt.collections_traversed)

        :param path: the URL path to match (usually this is ``PATH_INFO``)
        :param method: the HTTP method to match (usually this is
                       ``REQUEST_METHOD``). May be ``None``, in which case
                       matching will be performed on the ``path`` argument
                       only.
        :param request: a :class:`~fresco.request.Request` object

        :return: A generator yielding ``RouteTraversal`` objects
        """
        for route in self.__routes__:

            result = route.match(path, method)
            if result is None:
                continue

            if request and route.predicate and not route.predicate(request):
                continue

            # View function arguments extracted while traversing the path
            traversal_args = result.args
            traversal_kwargs = result.kwargs

            # Process any args/kwargs defined in the Route declaration.
            if request:
                route_args = tuple(
                    (item(request) if isinstance(item, RouteArg) else item)
                    for item in route.view_args)

                route_kwargs = dict(
                    (k, v(request) if isinstance(v, RouteArg) else v)
                    for k, v in route.view_kwargs.items())
            else:
                route_args = route.view_args
                route_kwargs = route.view_kwargs

            args = route_args + traversal_args
            kwargs = dict(traversal_kwargs, **route_kwargs)

            if isinstance(route, DelegateRoute):
                try:
                    sub_routes = route.routecollectionfactory(*args, **kwargs)
                except ResponseException as e:
                    def raiser(*args, **kwargs):
                        raise e
                    r = self.route_class('/', ALL_METHODS, raiser)
                    yield RouteTraversal(r, (), {}, [(self, '', r)])
                    continue

                for sub in sub_routes.get_routes(result.path_remaining,
                                                 method,
                                                 request):
                    traversed = [TraversedCollection(self, '', route,
                                                     args,
                                                     kwargs,
                                                     traversal_args,
                                                     traversal_kwargs)]
                    traversed.extend(
                        TraversedCollection(c, result.path_matched + p, r,
                                            a, k, ta, tk)
                        for c, p, r, a, k, ta, tk in sub.collections_traversed)

                    # Dynamic routes consume their arguments when creating the
                    # sub RouteCollection.
                    if route.dynamic:
                        yield RouteTraversal(
                            sub.route, sub.args, sub.kwargs, traversed)
                    else:
                        yield RouteTraversal(sub.route,
                                            args + sub.args,
                                            dict(kwargs, **sub.kwargs),
                                            traversed)
            else:
                yield RouteTraversal(
                    route, args, kwargs,
                    [TraversedCollection(self, '', route, args, kwargs,
                                         traversal_args, traversal_kwargs)])

    def route(self, pattern, methods=None, view=None, *args, **kwargs):
        """
        Match a URL pattern to a view function. Can be used as a function
        decorator, in which case the ``view`` parameter should not be passed.

        :param pattern: A string that can be compiled into a path pattern
        :param methods: A list of HTTP methods the view is bound to
        :param view:    The view function. If not specified a function
                        decorator will be returned.

        Other parameters are as for the :class:`Route` constructor.
        """
        # Catch the common error of not providing a valid method
        if methods and not isinstance(methods, Iterable):
            raise TypeError("HTTP methods must be specified as a string "
                            "or iterable")

        # Called as a decorator?
        if methods is not None and view is None:
            def route_decorator(func):
                self.add_route(self.route_class(pattern, methods, func, *args,
                                                **kwargs))
                return func
            return route_decorator

        route = self.route_class(pattern, methods, view, *args, **kwargs)
        self.add_route(route)
        return route

    def route_wsgi(self, path, wsgiapp, *args, **kwargs):
        """
        Route requests to ``path`` to the given WSGI application
        """
        # XXX: need to shift path_info to script_name
        from fresco import context

        def fresco_wsgi_view():
            fake_start_response = lambda status, headers, exc_info=None: None
            return Response.from_wsgi(wsgiapp,
                                      context.request.environ,
                                      fake_start_response)

        return self.route_all(path, ALL_METHODS, fresco_wsgi_view, *args,
                              **kwargs)

    def route_all(self, path, *args, **kwargs):
        """\
        Expose a view for all URLs starting with ``path``.

        :param path: the path prefix at which the view will be routed
        """
        return self.route(MatchAllURLsPattern(path), *args, **kwargs)

    def include(self, path, views):
        """
        Include a view collection at the given path.

        The included view collection's url properties will be modified to
        generate the prefixed URLs.

        :param path:  Path at which to include the views
        :param views: Any collection of views (must expose a ``__routes___``
                      attribute)
        """
        routes = list(r.add_prefix(path) for r in views.__routes__)
        for r in routes:
            if r.instance is None:
                r = r.bindto(views)
            self.add_route(r)

    def delegate(self, path, app, dynamic=False, *args, **kwargs):
        """\
        Delegate all requests under ``path`` to ``app``

        :param path: the path prefix at which the app will be available
        :param app: a FrescoApp instance
        """
        route = DelegateRoute(path, app, dynamic, *args, **kwargs)
        self.add_route(route)
        return route

    def replace(self, viewspec, newroute):
        """
        Replace the route(s) identified by ``viewspec`` with a new
        Route.

        :param viewspec: The routed view callable, or its fully qualified name
                         ('package.module.view_function'), or the name of a
                         named route
        :param newroute: The replacement route. This may be None, in which
                         case the route will be removed without replacement
        """
        route = self.routefor(viewspec)

        if newroute:
            position = self.__routes__.index(route)
            self.__routes__[position] = newroute
        else:
            self.__routes__.remove(route)

        for k, r in list(self.__routed_views__.items()):
            if r is route:
                del self.__routed_views__[k]

    def remove(self, viewspec):
        """
        Remove the route(s) identified by ``viewspec``

        :param viewspec: The routed view callable, or its fully qualified name
                         ('package.module.view_function'), or the name of a
                         named route
        """
        self.replace(viewspec, None)


class DelegateRoute(Route):
    """\
    A specialisation of Route that is used to delegate a path prefix to another
    route collection.
    """

    def __init__(self, prefix, view, dynamic=False,
                 *args, **kwargs):
        pattern = ExtensiblePattern(prefix, match_entire_path=False)
        self.dynamic = dynamic
        if self.dynamic:
            self.routecollectionfactory = self._dynamic_routecollectionfactory
        else:
            self.routecollectionfactory = self._static_routecollectionfactory
            if not isinstance(view, RouteCollection):
                view = RouteCollection(view.__routes__)
        super(DelegateRoute, self).__init__(pattern, ALL_METHODS,
                                            view, *args, **kwargs)

    def _dynamic_routecollectionfactory(self, *args, **kwargs):
        """\
        Return the RouteCollection responsible for paths under this route
        """

        routes = self.getdecoratedview(self.getview(GET))(*args, **kwargs)

        if isinstance(routes, RouteCollection):
            return routes

        return RouteCollection(r.bindto(routes) for r in routes.__routes__)

    def _static_routecollectionfactory(self, *args, **kwargs):
        return self.getview(GET)


def register_converter(name, registry=ExtensiblePattern):
    """
    Class decorator that registers a converter class for use with
    ExtensiblePattern.

    Example::

        >>> @register_converter('hex')
        ... class HexStringConverter(Converter):
        ...     pattern = r'[a-f0-9]+'
        ...
    """
    def register_converter(cls):
        registry.register_converter(name, cls)
        return cls
    return register_converter



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
"""\
Utilities for interfacing with WSGI
"""

from __future__ import absolute_import

from functools import partial
from io import BytesIO
import sys

from fresco.compat import unquote, urlparse

__all__ = ['environ_to_unicode',
           'unicode_to_environ',
           'environ_to_bytes',
           'bytes_to_environ',
           'StartResponseWrapper',
           'ClosingIterator',
           'make_environ',
           ]


#: The core keys expected in the WSGI environ dict, as defined by PEP3333
WSGI_KEYS = set(['REQUEST_METHOD',
                 'SCRIPT_NAME',
                 'PATH_INFO',
                 'QUERY_STRING',
                 'CONTENT_TYPE',
                 'CONTENT_LENGTH',
                 'SERVER_NAME',
                 'SERVER_PORT',
                 'SERVER_PROTOCOL'])


def with_docstring_from(src):
    def with_docstring_from(tgt):
        try:
            tgt.__doc__ = src.__doc__
        except AttributeError:
            tgt.func_doc = src.func_doc
        return tgt
    return with_docstring_from


def _environ_to_unicode_py2(s, enc='UTF-8'):
    """
    Decode a WSGI environ value to a unicode string
    """
    return s.decode(enc)


@with_docstring_from(_environ_to_unicode_py2)
def _environ_to_unicode_py3(s, enc='UTF-8'):
    return s.encode('iso-8859-1').decode(enc)


def _unicode_to_environ_py2(s, enc='UTF-8'):
    """
    Return a unicode string encoded for a WSGI environ value

    For python 2, this simply returns ``s`` encoded using the specified
    encoding

    For python 3 this returns a 'bytes-as-unicode' string:

    - encode ``s`` using the specified encoding (eg utf8)
    - decode the resulting byte string as latin-1
    """
    return s.encode(enc, 'surrogateescape')


@with_docstring_from(_unicode_to_environ_py2)
def _unicode_to_environ_py3(s, enc='UTF-8'):
    return s.encode(enc, 'surrogateescape').decode('iso-8859-1')


def _environ_to_bytes_py2(s):
    """
    Decode a WSGI environ value to a bytes object
    """
    return s


@with_docstring_from(_environ_to_bytes_py2)
def _environ_to_bytes_py3(s):
    return s.encode('latin1')


def _bytes_to_environ_py2(s):
    """
    Encode a byte string to a WSGI environ value

    For Python 2, this simply returns ``s``.
    For Python 3 this returns a 'bytes-as-unicode' string.
    """
    return s


@with_docstring_from(_environ_to_bytes_py2)
def _bytes_to_environ_py3(s):
    return s.decode('latin1')


if str is bytes:
    environ_to_unicode = _environ_to_unicode_py2
    unicode_to_environ = _unicode_to_environ_py2
    environ_to_bytes = _environ_to_bytes_py2
    bytes_to_environ = _bytes_to_environ_py2

else:
    environ_to_unicode = _environ_to_unicode_py3
    unicode_to_environ = _unicode_to_environ_py3
    environ_to_bytes = _environ_to_bytes_py3
    bytes_to_environ = _bytes_to_environ_py3


class StartResponseWrapper(object):
    """\
    Wrapper class for the ``start_response`` callable, allowing middleware
    applications to intercept and interrogate the proxied start_response
    arguments.

    Synopsis::

        >>> def my_wsgi_app(environ, start_response):
        ...     start_response('200 OK', [('Content-Type', 'text/plain')])
        ...     return ['Whoa nelly!']
        ...
        >>> def my_other_wsgi_app(environ, start_response):
        ...     responder = StartResponseWrapper(start_response)
        ...     result = my_wsgi_app(environ, responder)
        ...     print "Got status", responder.status
        ...     print "Got headers", responder.headers
        ...     responder.call_start_response()
        ...     return result
        ...
        >>> from flea import Agent
        >>> result = Agent(my_other_wsgi_app).get('/')
        Got status 200 OK
        Got headers [('Content-Type', 'text/plain')]

    See also ``Response.from_wsgi``, which takes a wsgi callable, environ and
    start_response, and returns a ``Response`` object, allowing the client to
    further interrogate and customize the WSGI response.

    Note that it is usually not advised to use this directly in middleware as
    start_response may not be called directly from the WSGI application, but
    rather from the iterator it returns. In this case the middleware may need
    logic to accommodate this. It is usually safer to use
    ``Response.from_wsgi``, which takes this into account.
    """

    def __init__(self, start_response):
        self.start_response = start_response
        self.status = None
        self.headers = []
        self.called = False
        self.buf = BytesIO()
        self.exc_info = None

    def __call__(self, status, headers, exc_info=None):
        """
        Dummy WSGI ``start_response`` function that stores the arguments for
        later use.
        """
        self.status = status
        self.headers = headers
        self.exc_info = exc_info
        self.called = True
        return self.buf.write

    def call_start_response(self):
        """
        Invoke the wrapped WSGI ``start_response`` function.
        """
        try:
            write = self.start_response(
                self.status,
                self.headers,
                self.exc_info
            )
            write(self.buf.getvalue())
            return write
        finally:
            # Avoid dangling circular ref
            self.exc_info = None


class ClosingIterator(object):
    """\
    Wrap a WSGI iterator to allow additional close functions to be called on
    application exit.

    Synopsis::

        >>> class filelikeobject(object):
        ...
        ...     def read(self):
        ...         print "file read!"
        ...         return ''
        ...
        ...     def close(self):
        ...         print "file closed!"
        ...
        >>> def app(environ, start_response):
        ...     f = filelikeobject()
        ...     start_response('200 OK', [('Content-Type', 'text/plain')])
        ...     return ClosingIterator(iter(f.read, ''), f.close)
        ...
        >>> from flea import Agent
        >>> m = Agent(app).get('/')
        file read!
        file closed!

    """

    def __init__(self, iterable, *close_funcs):
        """
        Initialize a ``ClosingIterator`` to wrap iterable ``iterable``, and
        call any functions listed in ``*close_funcs`` on the instance's
        ``.close`` method.
        """
        self.__dict__['_iterable'] = iterable
        self.__dict__['_next'] = partial(next, iter(self._iterable))
        self.__dict__['_close_funcs'] = close_funcs
        iterable_close = getattr(self._iterable, 'close', None)
        if iterable_close is not None:
            self.__dict__['_close_funcs'] = (iterable_close,) + close_funcs
        self.__dict__['_closed'] = False

    def __iter__(self):
        """
        ``__iter__`` method
        """
        return self

    def __next__(self):
        """\
        Return the next item from the iterator
        """
        return self._next()

    def next(self):
        """
        Return the next item from ``iterator``
        """
        return self._next()

    def close(self):
        """
        Call all close functions listed in ``*close_funcs``.
        """
        self.__dict__['_closed'] = True
        for func in self._close_funcs:
            func()

    def __getattr__(self, attr):
        return getattr(self._iterable, attr)

    def __setattr__(self, attr, value):
        return getattr(self._iterable, attr, value)

    def __del__(self):
        """
        Emit a warning if the iterator is deleted without ``close`` having been
        called.
        """
        if not self._closed:
            try:
                import warnings
            except ImportError:
                return
            else:
                warnings.warn("%r deleted without close being called" %
                              (self,))


class WSGIStack(object):
    """
    A context manager that sets up a WSGI request environ
    passed through a list of middleware.
    """

    def __init__(self, environ, middleware_stack):
        self.environ = environ
        self.middleware_stack = middleware_stack
        app = self.fake_app
        for m in middleware_stack:
            app = m(app)
        self.app = app
        self.content_iterator = self.app(environ, lambda *a, **kw: None)

    def fake_app(self, environ, start_response):
        start_response('200 OK', [])
        yield b''

    def __enter__(self):
        # Invoke the fake WSGI app. This executes any middleware, which may
        # have side effects required downstream (eg setting environ keys)
        next(self.content_iterator, None)
        return self.environ

    def __exit__(self, type, value, traceback):
        # Exhaust the iterator and close up
        list(self.content_iterator)
        getattr(self.content_iterator, 'close', lambda: None)()


def make_environ(url='/', environ=None,
                    wsgi_input=b'', middleware=True, **kwargs):
    """
    Return a WSGI environ dict populated with values modelling the specified
    request url and data.

    :param url: The URL for the request,
                eg ``/index.html`` or ``/search?q=foo``.
                Note that ``url`` must be properly URL encoded.
    :param environ: values to pass into the WSGI environ dict
    :param wsgi_input: The input stream to use in the ``wsgi.input``
                        key of the environ dict
    :param middleware: If ``False`` the middleware stack will not be
                        invoked. Disabling the middleware can speed up
                        the execution considerably, but it will no longer
                        give an accurate simulation of a real HTTP request.
    :param kwargs: additional keyword arguments will be passed into the
                    WSGI request environment
    """
    from fresco.request import HEADER_NAMES as REQUEST_HEADER_NAMES
    url = urlparse(url)
    netloc = url.netloc
    user = ''
    if '@' in netloc:
        user, netloc = netloc.split('@', 1)

    if ':' in user:
        user, _ = user.split(':')[0]

    if isinstance(wsgi_input, bytes):
        wsgi_input = BytesIO(wsgi_input)

    env_overrides = environ or {}
    for key, value in kwargs.items():
        key = key.replace('wsgi_', 'wsgi.')
        if '.' not in key:
            # Convert core WSGI keys to upper case
            if key.upper() in WSGI_KEYS:
                key = key.upper()
            # Convert header names to form HTTP_USER_AGENT
            elif key.lower() in REQUEST_HEADER_NAMES:
                key = 'HTTP_' + key.upper()
            # value must be a python native str, whatever that means
            if not isinstance(value, str):
                value = unicode_to_environ(value)
        env_overrides[key] = value

    environ = {
        'REQUEST_METHOD': 'GET',
        'SERVER_NAME': 'localhost',
        'SERVER_PORT': '80',
        'SERVER_PROTOCOL': 'HTTP/1.0',
        'HTTP_HOST': netloc or 'localhost',
        'SCRIPT_NAME': '',
        'PATH_INFO': unicode_to_environ(unquote(url.path)),
        'REMOTE_ADDR': '127.0.0.1',
        'wsgi.version': (1, 0),
        'wsgi.url_scheme': url.scheme or 'http',
        'wsgi.input': wsgi_input,
        'wsgi.errors': sys.stderr,
        'wsgi.multithread': True,
        'wsgi.multiprocess': True,
        'wsgi.run_once': False,
    }
    if url.scheme == 'https':
        environ['HTTPS'] = 'on'
        environ['SERVER_PORT'] = '443'

    if user:
        environ['REMOTE_USER'] = user

    if url.query:
        environ['QUERY_STRING'] = url.query

    environ.update(env_overrides)
    return environ



# encoding: utf-8
# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
from __future__ import absolute_import

from copy import copy
from functools import wraps

from mock import Mock, call
import pytest
import tms

from fresco import FrescoApp, ALL_METHODS, GET, POST, Response
from fresco.exceptions import NotFound
from fresco.compat import ustr
from fresco.core import urlfor
from fresco.routing import (
    Route, DelegateRoute, RouteCollection, routefor, RouteTraversal,
    RouteNotFound, TraversedCollection)
from . import fixtures


def assert_method_bound_to(method, ob):
    try:
        assert method.__self__ is ob
    except AttributeError:
        assert method.im_self is ob


class TestMethodDispatch(object):

    def test_route_is_dispatched_to_correct_method(self):

        getview = Mock(return_value=Response())
        postview = Mock(return_value=Response())
        app = FrescoApp()
        app.route('/', GET, getview)
        app.route('/', POST, postview)

        with app.requestcontext('/'):
            app.view()
            assert getview.call_count == 1
            assert postview.call_count == 0

        with app.requestcontext_post('/'):
            app.view()
            assert getview.call_count == 1
            assert postview.call_count == 1


class TestRouteConstructor(object):

    def test_multiple_views_can_be_associated_with_a_route(self):

        app = FrescoApp()
        v1 = Mock(return_value=Response())
        v2 = Mock(return_value=Response())
        app.route('/', GET=v1, POST=v2)

        with app.requestcontext():
            app.view()
            assert v1.call_count == 1
            assert v2.call_count == 0

        with app.requestcontext_post():
            app.view()
            assert v1.call_count == 1
            assert v2.call_count == 1

    def test_kwargs_take_precedence(self):
        app = FrescoApp()
        v1 = Mock(return_value=Response())
        v2 = Mock(return_value=Response())
        app.route('/', ALL_METHODS, v1, POST=v2)

        with app.requestcontext():
            app.view()
            assert v1.call_count == 1
            assert v2.call_count == 0

        with app.requestcontext_post():
            app.view()
            assert v1.call_count == 1
            assert v2.call_count == 1

    def test_it_catches_invalid_methods(self):

        # Route must be a string or an iterable
        with pytest.raises(TypeError):
            Route('/', object(), lambda: None)

        # Route must be a valid HTTP method
        with pytest.raises(ValueError):
            Route('/', 'FOO', lambda: None)

        with pytest.raises(ValueError):
            Route('/', ['GET', 'FOO'], lambda: None)


class TestRouteViewFilters(object):

    def exclaim(self, response):
        return response.replace(
            content=[b''.join(response.content_iterator) + b'!'])

    def ask(self, response):
        return response.replace(
            content=[b''.join(response.content_iterator) + b'?'])

    def test_filter_is_applied(self):
        views = fixtures.CBV('test')
        app = FrescoApp()
        app.route('/', GET, views.index_html).filter(self.ask)
        with app.requestcontext('/'):
            assert app.view().content_iterator == [b'test?']

    def test_filter_is_applied_as_route_kwargs(self):
        views = fixtures.CBV('test')
        app = FrescoApp()
        app.route('/', GET, views.index_html, filters=[self.ask])
        with app.requestcontext('/'):
            assert app.view().content_iterator == [b'test?']

    def test_it_applies_multiple_filters_in_order(self):
        app = FrescoApp()
        views = fixtures.CBV('test')
        app.route('/', GET, views.index_html).filter(self.ask, self.exclaim)
        with app.requestcontext('/'):
            assert app.view().content_iterator == [b'test?!']

    def test_it_applies_chained_filter_calls_in_order(self):
        app = FrescoApp()
        views = fixtures.CBV('test')
        app.route('/', GET, views.index_html)\
                .filter(self.ask).filter(self.exclaim)
        with app.requestcontext('/'):
            assert app.view().content_iterator == [b'test?!']


class TestRouteDecorators(object):

    def exclaim(self, func):
        @wraps(func)
        def exclaim(*args, **kwargs):
            response = func(*args, **kwargs)
            return response.replace(
                content=[b''.join(response.content_iterator) + b'!'])
        return exclaim

    def test_decorator_is_applied(self):

        views = fixtures.CBV('test')

        app = FrescoApp()
        app.route('/decorated', GET, views.index_html,
                  decorators=[self.exclaim])
        app.route('/plain', GET, views.index_html)

        with app.requestcontext('/decorated'):
            assert app.view().content_iterator == [b'test!']

        with app.requestcontext('/plain'):
            assert app.view().content_iterator == [b'test']

    def test_decorator_is_applied_with_wrap_method(self):

        views = fixtures.CBV('test')

        app = FrescoApp()
        app.route('/decorated', GET, views.index_html).wrap(self.exclaim)
        app.route('/plain', GET, views.index_html)

        with app.requestcontext('/decorated'):
            assert app.view().content_iterator == [b'test!']

        with app.requestcontext('/plain'):
            assert app.view().content_iterator == [b'test']

    def test_decorator_works_with_urlfor(self):

        views = fixtures.CBV('test')
        app = FrescoApp()
        app.route('/decorated', GET, views.index_html,
                  decorators=[self.exclaim])
        with app.requestcontext():
            assert urlfor(views.index_html, _app=app) == \
                'http://localhost/decorated'

    def test_using_wraps_with_viewspec_doesnt_raise_AttributeError(self):

        def decorator(func):
            @wraps(func)
            def decorator(*args, **kwargs):
                return func(*args, **kwargs)
            return decorator

        class Views(object):
            __routes__ = Route('/',
                               GET, 'index_html', decorators=[decorator]),

            def index_html(self):
                return Response(['hello'])

        app = FrescoApp()
        app.include('/', Views())


class TestPredicates(object):

    def test_predicate_match(self):

        def v1():
            return Response([b'x'])

        def v2():
            return Response([b'y'])

        app = FrescoApp()
        app.route('/', GET, v1, predicate=lambda request: 'x' in request.query)
        app.route('/', GET, v2, predicate=lambda request: 'y' in request.query)

        with app.requestcontext('/?x=1'):
            assert b''.join(app.view().content) == b'x'
        with app.requestcontext('/?y=1'):
            assert b''.join(app.view().content) == b'y'
        with app.requestcontext('/'):
            assert app.view().status_code == 404


class TestRouteNames(object):

    def test_name_present_in_route_keys(self):
        r = Route('/', GET, None, name='foo')
        assert 'foo' in list(r.route_keys())

    def test_name_with_other_kwargs(self):
        r = Route('/', GET, None, name='foo', x='bar')
        assert 'foo' in list(r.route_keys())

    def test_name_cannot_contain_colon(self):
        with pytest.raises(ValueError):
            Route('/', GET, None, name='foo:bar')


class TestRouteCollection(object):

    def test_it_adds_routes_from_constructor(self):
        r1 = Route('/1', GET, None, name='1')
        r2 = Route('/2', POST, None, name='2')
        rc = RouteCollection([r1, r2])
        assert [r.name for r in rc] == ['1', '2']

    def test_it_adds_routecollections_from_constructor(self):
        r1 = Route('/', GET, None, name='1')
        r2 = Route('/', POST, None, name='2')
        r3 = Route('/', POST, None, name='3')
        rc = RouteCollection([r1, RouteCollection([r2, r3])])
        assert [r.name for r in rc] == ['1', '2', '3']

    def test_it_adds_dunderroutes_from_constructor(self):
        r1 = Route('/', GET, None, name='1')
        r2 = Route('/', POST, None, name='2')
        r3 = Route('/', POST, None, name='3')

        class A:
            __routes__ = [r2, r3]

        rc = RouteCollection([r1, A()])
        assert [r.name for r in rc] == ['1', '2', '3']

    def test_get_routes_matches_on_method(self):

        r_get = Route('/', GET, None)
        r_post = Route('/', POST, None)

        rc = RouteCollection([r_post, r_get])

        assert [r.route for r in rc.get_routes('/', GET)] == [r_get]
        assert [r.route for r in rc.get_routes('/', POST)] == [r_post]

    def test_get_routes_matches_on_path(self):

        r1 = Route('/1', GET, None)
        r2 = Route('/2', GET, None)

        rc = RouteCollection([r1, r2])

        assert [r.route for r in rc.get_routes('/1', GET)] == [r1]
        assert [r.route for r in rc.get_routes('/2', GET)] == [r2]

    def test_get_routes_can_match_all_methods(self):

        r1 = Route('/1', GET, None)
        r2 = Route('/1', POST, None)

        rc = RouteCollection([r1, r2])
        assert [r.route for r in rc.get_routes('/1', None)] == [r1, r2]

    def test_route_returns_traversal_information_on_nested_routes(self):

        a = RouteCollection()
        b = RouteCollection()

        a_route = Route('/harvey', GET, lambda: None)
        b_route = Route('/harvey', GET, lambda: None)

        a.add_route(a_route)
        b.add_route(b_route)

        a_delegate_route = DelegateRoute('/rabbit', b)
        b_delegate_route = DelegateRoute('/hole', a)

        a.add_route(a_delegate_route)
        b.add_route(b_delegate_route)

        r = next(a.get_routes('/rabbit/hole/rabbit/harvey', None))

        assert r.collections_traversed == [
            (a, '', a_delegate_route, (), {}, (), {}),
            (b, '/rabbit', b_delegate_route, (), {}, (), {}),
            (a, '/rabbit/hole', a_delegate_route, (), {}, (), {}),
            (b, '/rabbit/hole/rabbit', b_route, (), {}, (), {})]

    def test_pathfor_works_with_positional_args(self):
        view = Mock(return_value=Response())
        rc = RouteCollection([Route('/<:str>', GET, view)])
        assert rc.pathfor(view, 'x') == '/x'

    def test_replace_raises_route_not_found(self):
        a = RouteCollection()
        view = Mock(return_value=Response())
        a.route('/harvey', GET, view, name='harvey')
        with pytest.raises(RouteNotFound):
            a.replace('rabbit', None)

    def test_replace_selects_routes_by_name(self):
        a = RouteCollection()
        oldroute = Route('/', GET, Mock(), name='harvey')
        newroute = Route('/', GET, Mock())
        a.add_route(oldroute)
        a.replace('harvey', newroute)
        assert a.__routes__ == [newroute]

    def test_replace_selects_routes_by_view(self):
        a = RouteCollection()
        view = Mock(return_value=Response())
        oldroute = Route('/', GET, view)
        newroute = Route('/', GET, Mock())
        a.add_route(oldroute)
        a.replace(view, newroute)
        assert a.__routes__ == [newroute]

    def test_can_add_a_list_to_a_routecollection(self):
        r1 = Route('/', GET, Mock())
        r2 = Route('/', GET, Mock())
        assert (RouteCollection([r1]) + [r2]).__routes__ == [r1, r2]

    def test_can_add_route_to_routecollection(self):
        r1 = Route('/', GET, Mock())
        r2 = Route('/', GET, Mock())
        assert (RouteCollection([r1]) + r2).__routes__ == [r1, r2]

    def test_can_add_routecollection_to_route(self):
        r1 = Route('/', GET, Mock())
        r2 = Route('/', GET, Mock())
        assert (r1 + RouteCollection([r2])).__routes__ == [r1, r2]

    def test_can_add_routecollections(self):
        r1 = Route('/', GET, Mock())
        r2 = Route('/', GET, Mock())
        assert (RouteCollection([r1]) + RouteCollection([r2])).__routes__ == \
                [r1, r2]

    def test_routecollections_can_be_used_in_classes(self):
        class MyViews(object):
            __routes__ = RouteCollection([
                Route('/', GET, 'view')])

            def view(self):
                return Response()

        v = MyViews()
        app = FrescoApp()
        app.include('/', v)
        assert [r.route.getview(GET)
                for r in app.get_routes('/', GET)] == [v.view]

    def test_routecollections_in_classes_can_be_manipulated(self):
        class MyViews(object):
            __routes__ = RouteCollection([
                Route('/', GET, 'view')])

            def view(self):
                return Response()

        class MyOtherViews(MyViews):
            __routes__ = copy(MyViews.__routes__)
            __routes__.replace('view', Route('/', GET, 'another_view'))

            def another_view(self):
                return Response()

        v = MyOtherViews()
        app = FrescoApp()
        app.include('/', v)
        assert [r.route.getview(GET)
                for r in app.get_routes('/', GET)] == [v.another_view]

    def test_add_prefix_returns_prefixed_collection(self):
        rc = RouteCollection([Route('/fish', GET, None),
                              Route('/beans', GET, None)])
        prefixed = rc.add_prefix('/jelly')
        assert [str(r.pattern) for r in prefixed] == \
                ['/jelly/fish', '/jelly/beans']

    def test_it_binds_routes_to_an_instance(self):
        views = fixtures.CBV('test')
        rc = RouteCollection([views])
        view = next(rc.get_routes('/', GET)).route.getview(GET)
        assert_method_bound_to(view, views)

    def test_it_binds_routes_to_an_instance_via_include(self):
        views = fixtures.CBV('test')
        rc = RouteCollection([])
        rc.include('/', views)
        view = next(rc.get_routes('/', GET)).route.getview(GET)
        assert_method_bound_to(view, views)

    def test_including_twice_does_not_rebind_instance(self):
        views = fixtures.CBV('test')
        rc = RouteCollection([])
        rc.include('/', views)
        rc2 = RouteCollection([])
        rc2.include('/', rc)
        view = next(rc2.get_routes('/', GET)).route.getview(GET)
        assert_method_bound_to(view, views)


class TestRoutefor(object):

    def test_routefor_with_view_function(self):

        def view():
            return Response(['ok'])

        app = FrescoApp()
        route = app.route('/foo', GET, view)

        with app.requestcontext():
            assert routefor(view) == route

    def test_routefor_with_string(self):
        app = FrescoApp()
        route = app.route('/myviewfunc', GET, fixtures.module_level_function)
        with app.requestcontext():
            assert routefor('fresco.tests.fixtures.module_level_function') == \
                route

    def test_routefor_generates_first_route(self):

        myviewfunc = lambda req: Response([])
        app = FrescoApp()
        r1 = app.route('/1', GET, myviewfunc)
        app.route('/2', GET, myviewfunc)
        with app.requestcontext():
            assert routefor(myviewfunc) == r1


class TestDelegatedRoutes(object):

    def test_dispatch_to_delegated_route(self):

        def hello():
            return Response([b'hello'])

        inner = FrescoApp()
        inner.route('/hello', GET, hello)

        outer = FrescoApp()
        outer.delegate('/say', inner)

        with outer.requestcontext('/say/hello'):
            assert b''.join(outer.view().content) == b'hello'

    def test_url_variables_are_passed(self):

        hello = Mock(return_value=Response())

        inner = FrescoApp()
        inner.route('/<i:str>', GET, hello)

        outer = FrescoApp()
        outer.delegate('/<o:str>', inner)

        with outer.requestcontext('/foo/bar'):
            outer.view()
            assert hello.call_args_list == [call(i='bar', o='foo')]

    def test_delegation_to_dynamic_routes(self):

        result = []

        class MyRoutes(object):
            __routes__ = [Route('/<inner:int>', GET, 'view')]

            def __init__(self, **kwargs):
                self.kwargs = kwargs

            def view(self, **kwargs):
                result.append((self, kwargs))
                return Response()

        app = FrescoApp()
        app.delegate('/<outer:str>', MyRoutes, dynamic=True)
        with app.requestcontext('/one/2'):
            app.view()
            instance, inner_kwargs = result[0]
            assert instance.kwargs == {'outer': 'one'}
            assert inner_kwargs == {'inner': 2}

    def test_dynamic_routes_are_never_shared(self):

        result = []

        class MyRoutes(object):
            __routes__ = [Route('', GET, 'view')]

            def __init__(self, value):
                self.value = value

            def view(self):
                result.append(self.value)
                return Response()

        app = FrescoApp()
        app.delegate('/<value:str>', MyRoutes, dynamic=True)
        with app.requestcontext('/one'):
            app.view()
            v1 = result.pop()
        with app.requestcontext('/two'):
            app.view()
            v2 = result.pop()
        assert v1 == 'one'
        assert v2 == 'two', v2

    def test_pathfor_with_delegated_route(self):
        inner = FrescoApp()
        inner.route('/<i:str>', GET, lambda: None, name='inner-route')

        outer = FrescoApp()
        outer.delegate('/<o:str>', inner, name='delegation')

        with outer.requestcontext('/foo/bar'):
            assert outer.pathfor('delegation:inner-route',
                                 o='x', i='y') == '/x/y'

    def test_pathfor_with_dynamic_delegated_route(self):

        view = Mock(return_value=Response())

        def routecollectionfactory(*args, **kwargs):
            return RouteCollection([Route('/<i:str>', GET, view,
                                          name='inner-route')])

        rc = RouteCollection()
        rc.delegate('/<o:str>', routecollectionfactory,
                    name='delegation', dynamic=True)

        assert rc.pathfor('delegation:inner-route', o='x', i='y') == '/x/y'

    def test_pathfor_with_dynamic_delegated_route_uses_default_args(self):

        view = Mock(return_value=Response())

        def routecollectionfactory(factoryarg1, factoryarg2):
            return RouteCollection([Route('/<i:str>', GET, view,
                                          name='inner-route')])

        rc = RouteCollection()
        rc.delegate('/<factoryarg1:str>/<factoryarg2:str>',
                    routecollectionfactory,
                    factoryarg1_default='foo',
                    factoryarg2_default=lambda r: 'bar',
                    name='delegation', dynamic=True)

        assert rc.pathfor('delegation:inner-route', i='y') == '/foo/bar/y'

    def test_urlfor_with_dynamic_delegated_route_and_view_self(self):

        result = []

        class MyRoutes(object):
            __routes__ = [Route('/<inner:int>/view', GET, 'view')]

            def __init__(self, **kwargs):
                self.kwargs = kwargs

            def view(self, **kwargs):
                result.append(urlfor(self.view, inner=3))
                return Response()

        app = FrescoApp()
        app.delegate('/<outer:str>', MyRoutes, dynamic=True)
        with app.requestcontext('/two/2/view'):
            app.view()
            assert result == ['http://localhost/two/3/view']

    def test_urlgeneration_with_dynamic_routes(self):

        class Routable(object):
            __routes__ = [Route('/<b:int>', GET, 'view', name='y')]

            def __init__(self, a):
                pass

            def view(self, b):
                return Response()

        app = FrescoApp()
        app.delegate('/<a:str>', Routable, dynamic=True, name='x')
        with app.requestcontext('/two/2/view'):
            assert urlfor('x:y', a='a', b=1) == 'http://localhost/a/1'

    def test_delegated_routes_can_be_included(self):

        view = Mock(return_value=Response())

        inner = RouteCollection([Route('/baz', GET, view)])
        middle = RouteCollection([DelegateRoute('/bar', inner)])
        outer = FrescoApp()
        outer.include('/foo', middle)
        with outer.requestcontext('/foo/bar/baz'):
            outer.view()
            assert view.call_count == 1

    def test_not_found_is_returned(self):

        def inner():
            raise NotFound()

        outer = FrescoApp()
        outer.delegate('/foo', inner, dynamic=True)
        with outer.requestcontext('/foo/bar/baz'):
            response = outer.view()
            assert response.status_code == 404

    def test_not_found_causes_next_route_to_be_tried(self):

        def inner():
            raise NotFound()
        view = Mock(return_value=Response())

        outer = FrescoApp()
        outer.delegate('/foo', inner, dynamic=True)
        outer.route('/foo', GET, view)
        with outer.requestcontext('/foo'):
            outer.view()
            assert view.call_count == 1


class TestConverters(object):

    def test_str_converter_returns_unicode(self):
        from fresco.routing import StrConverter
        s = ustr('abc')
        assert isinstance(StrConverter().from_string(s), ustr)

    def test_register_coverter_acts_as_decorator(self):
        from fresco.routing import Converter, register_converter

        @register_converter('testconverter')
        class MyConverter(Converter):
            def from_string(self, s):
                return 'bar'

        view = Mock(return_value=Response())

        app = FrescoApp()
        app.route('/<:testconverter>', GET, view)
        with app.requestcontext('/foo'):
            app.view()
            assert view.call_args == (('bar',), {}), view.call_args


class TestViewArgs(object):

    def test_it_uses_args(self):
        routes = RouteCollection([Route('/', GET, None, args=(1, 2))])
        assert list(routes.get_routes('/', GET)) == \
                [tms.InstanceOf(RouteTraversal, args=(1, 2))]

    def test_it_uses_view_args(self):
        routes = RouteCollection([Route('/', GET, None, view_args=(1, 2))])
        assert list(routes.get_routes('/', GET)) == \
                [tms.InstanceOf(RouteTraversal, args=(1, 2))]

    def test_it_appends_args_extracted_from_path(self):
        routes = RouteCollection([Route('/<:int>', GET, None,
                                        view_args=(1, 2))])
        assert list(routes.get_routes('/3', GET)) == \
                [tms.InstanceOf(RouteTraversal, args=(1, 2, 3))]

    def test_it_keeps_traversal_args_separate(self):
        routes = RouteCollection([Route('/<:int>', GET, None, view_args=(1,))])
        assert list(routes.get_routes('/2', GET)) == \
            [tms.InstanceOf(RouteTraversal, args=(1, 2),
                            collections_traversed=[
                                tms.InstanceOf(TraversedCollection,
                                               args=(1, 2),
                                               traversal_args=(2,))])]


class TestViewKwargs(object):

    def test_it_reads_from_route_kwargs(self):
        routes = RouteCollection([Route('/', GET, None, x=1)])
        assert list(routes.get_routes('/', GET)) == \
                [tms.InstanceOf(RouteTraversal, kwargs={'x': 1})]

    def test_it_reads_from_kwargs(self):
        routes = RouteCollection([Route('/', GET, None, kwargs={'x': 1})])
        assert list(routes.get_routes('/', GET)) == \
                [tms.InstanceOf(RouteTraversal, kwargs={'x': 1})]

    def test_it_reads_from_view_kwargs(self):
        routes = RouteCollection([Route('/', GET, None, view_kwargs={'x': 1})])
        assert list(routes.get_routes('/', GET)) == \
                [tms.InstanceOf(RouteTraversal, kwargs={'x': 1})]

    def test_it_keeps_traversal_kwargs_separate(self):
        routes = RouteCollection([
            Route('/<x:int>', GET, None, view_kwargs={'y': 1})])
        assert list(routes.get_routes('/2', GET)) == \
            [tms.InstanceOf(RouteTraversal,
                            kwargs={'x': 2, 'y': 1},
                            collections_traversed=[
                                tms.InstanceOf(TraversedCollection,
                                               kwargs={'y': 1, 'x': 2},
                                               traversal_kwargs={'x': 2})])]


class TestRouteClassIsPluggable(object):

    class CustomRoute(Route):
        pass

    def test_it_defaults_to_Route(self):
        routes = RouteCollection()
        assert routes.route_class is Route

    def test_it_accepts_route_class_arg(self):
        routes = RouteCollection(route_class=self.CustomRoute)
        assert routes.route_class is self.CustomRoute

    def test_it_uses_route_class_in_route_method(self):

        def myview():
            pass

        routes = RouteCollection(route_class=self.CustomRoute)
        routes.route('/', GET, myview)

        assert list(routes.get_routes('/', GET)) == \
                [tms.InstanceOf(RouteTraversal,
                                route=tms.InstanceOf(self.CustomRoute))]

    def test_it_uses_route_class_in_decorator(self):

        routes = RouteCollection(route_class=self.CustomRoute)

        @routes.route('/', GET)
        def myview():
            pass

        assert list(routes.get_routes('/', GET)) == \
                [tms.InstanceOf(RouteTraversal,
                                route=tms.InstanceOf(self.CustomRoute))]

    def test_custom_route_class_survives_include(self):
        routes = RouteCollection(route_class=self.CustomRoute)

        @routes.route('/', GET)
        def myview():
            pass

        routes2 = RouteCollection()
        routes2.include('/incl', routes)

        assert list(routes2.get_routes('/incl/', GET)) == \
                [tms.InstanceOf(RouteTraversal,
                                route=tms.InstanceOf(self.CustomRoute))]


class TestRouteTraversal(object):

    routes = RouteCollection()
    inner = RouteCollection()
    inner2 = RouteCollection()

    view = object()
    inner2.route('/<i:int>', GET=view, name='c')
    inner.delegate('/<i:int>', inner2, name='b')
    routes.delegate('/<i:int>', lambda i, r=inner: r, dynamic=True, name='a')

    def test_it_reconstructs_the_path(self):
        traversal = next(self.routes.get_routes('/1/2/3', GET))

        # Do we get the original path back?
        assert traversal.build_path() == '/1/2/3'

    def test_it_modifies_the_path(self):

        traversal = next(self.routes.get_routes('/1/2/3', GET))

        # Modify the traversal and check we get a modified path back
        assert traversal.replace('a', {'i': 0}).build_path() == '/0/2/3'

        assert traversal.replace('b', {'i': 0}).build_path() == '/1/0/3'
        assert traversal.replace('a:b', {'i': 0}).build_path() == '/1/0/3'

        assert traversal.replace('a:b:c', {'i': 0}).build_path() == '/1/2/0'
        assert traversal.replace('a:c', {'i': 0}).build_path() == '/1/2/0'
        assert traversal.replace('c', {'i': 0}).build_path() == '/1/2/0'
        assert traversal.replace(self.view, {'i': 0}).build_path() == '/1/2/0'

        # Passing no arguments should create a new traversal
        assert traversal.replace('a') is not traversal

        # ...identical to the original
        assert traversal.replace('a') == traversal

    def test_replace_raises_routenotfound(self):
        traversal = next(self.routes.get_routes('/1/2/3', GET))
        with pytest.raises(RouteNotFound):
            traversal.replace('x')

        with pytest.raises(RouteNotFound):
            traversal.replace('a:x')

        with pytest.raises(RouteNotFound):
            traversal.replace('b:a')

        with pytest.raises(RouteNotFound):
            traversal.replace('a:b:c:d')

        with pytest.raises(RouteNotFound):
            traversal.replace('')



# encoding=UTF-8
# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
# See LICENSE.txt for terms of redistribution and use.

from __future__ import absolute_import
from __future__ import unicode_literals
from io import BytesIO

from fresco.compat import quote, ustr
from fresco.util.http import parse_querystring, parse_post, encode_multipart
from fresco import FrescoApp, context, Response, POST

from . import form_data


class TestParseQueryString(object):

    def p(self, value):
        return list(parse_querystring(value))

    def test_empty(self):
        self.p('') == []

    def test_simple_key_value(self):
        assert self.p('a=b') == [('a', 'b')]

    def test_key_with_space(self):
        assert self.p('a+b=c') == [('a b', 'c')]

    def test_value_with_space(self):
        assert self.p('a=b+c') == [('a', 'b c')]

    def test_double_equals(self):
        assert self.p('a==b') == [('a', '=b')]

    def test_escaped_chars(self):
        assert self.p('%20==c%3D') == [(' ', '=c=')]

    def test_charset(self):
        assert self.p('a=el%20ni%C3%B1o') == [('a', 'el niño')]


class TestParseMultipart(object):

    def test_multipart(self):
        for data in form_data.multipart_samples:
            io = BytesIO(data['data'])
            io.seek(0)
            environ = {
                'CONTENT_LENGTH': data['content_length'],
                'CONTENT_TYPE': data['content_type'],
            }
            parsed = sorted(list(parse_post(environ, io, 'UTF-8')))

            assert [name for name, value in parsed] == \
["empty-text-input", "file-upload", "text-input-ascii", "text-input-unicode"]

            assert parsed[0] == ("empty-text-input", "")
            assert parsed[2] == ("text-input-ascii", "abcdef")
            assert parsed[3] == ("text-input-unicode",
                                    b"\xce\xb1\xce\xb2\xce\xb3\xce\xb4"
                                    .decode("utf8"))

            fieldname, fileupload = parsed[1]
            assert fieldname == "file-upload"
            assert fileupload.filename == "test.data"
            assert fileupload.headers['content-type'] == \
                        "application/octet-stream"
            assert fileupload.file.read() == form_data.FILE_UPLOAD_DATA

    def test_fileupload_too_big(self):
        """\
        Verify that multipart/form-data encoded POST data raises an exception
        if the total data size exceeds request.MAX_SIZE bytes
        """

        def view():
            request = context.request
            request.MAX_MULTIPART_SIZE = 500
            request.get('f1')
            return Response(['ok'])
        app = FrescoApp()
        app.route('/', POST, view)

        with app.requestcontext_post(
                '/', files=[('f1', 'filename.txt', 'text/plain', 'x' * 1000)]):
            assert app.view().status == "413 Payload Too Large"

        with app.requestcontext_post(
                '/',
                files=[('f1', 'filename.txt', 'text/plain', 'x' * 400)],
                data={'f2': 'x' * 101}):
            assert app.view().status == "413 Payload Too Large"

    def test_fileupload_with_invalid_content_length(self):

        def view():
            request = context.request
            request.get('f1')
            return Response(['ok'])
        app = FrescoApp()
        app.route('/', POST, view)

        with app.requestcontext_post(
                '/',
                files=[('f1', 'filename.txt', 'text/plain', 'x' * 1000)]) as c:
            c.request.environ['CONTENT_LENGTH'] = str('500')
            assert app.view().status == "400 Bad Request"

    def test_multipart_field_too_big(self):
        """
        Verify that multipart/form-data encoded POST data raises an exception
        if it contains a single field exceeding request.MAX_SIZE bytes
        """
        def view():
            request = context.request
            request.MAX_MULTIPART_SIZE = 500
            request.MAX_SIZE = 100
            request.get('f1')
            return Response(['ok'])
        app = FrescoApp()
        app.route('/', POST, view)

        with app.requestcontext_post('/',
                                     multipart=True,
                                     data=[('f1', 'x' * 200)]):
            assert app.view().status == "413 Payload Too Large"


class TestParseFormEncodedData(object):

    char_latin1 = b'\xa3'
    char_utf8 = b'\xc2\xa3'
    char = char_latin1.decode('latin1')

    assert char_latin1.decode('latin1') == char_utf8.decode('utf8') == char

    def test_formencoded_data_too_big(self):
        """
        Verify that application/x-www-form-urlencoded POST data raises an
        exception if it exceeds request.MAX_SIZE bytes
        """
        def view():
            request = context.request
            request.MAX_SIZE = 100
            request.get('f1')
            return Response(['ok'])
        app = FrescoApp()
        app.route('/', POST, view)

        with app.requestcontext_post('/',
                                     data=[('f1', 'x' * 200)]):
            assert app.view().status == "413 Payload Too Large"

    def test_posted_data_contains_non_ascii_chars(self):
        """
        Verify that it raises an exception if POST data contains invalid
        characters.
        """
        def view():
            context.request.get('f1')

        app = FrescoApp()
        app.route('/', POST, view)

        data = 'foo=bár'.encode('utf8')

        with app.requestcontext(
                '/',
                REQUEST_METHOD='POST',
                CONTENT_LENGTH=ustr(len(data)),
                CONTENT_TYPE='application/x-www-form-urlencoded',
                wsgi_input=data):
            response = app.view()
            assert response.status == "400 Bad Request"

    def test_non_utf8_data_posted(self):

        data = b'char=' + quote(self.char_latin1).encode('ascii')
        env = {'REQUEST_METHOD': 'POST',
               'CONTENT_LENGTH': str(len(data)),
               'wsgi.input': BytesIO(data)}

        with FrescoApp().requestcontext(environ=env) as c:
            request = c.request
            request.charset = 'latin1'
            assert request.form['char'] == self.char

    def test_non_utf8_data_getted(self):

        data = 'char=' + quote(self.char_latin1)
        env = {'REQUEST_METHOD': 'GET', 'QUERY_STRING': data}

        with FrescoApp().requestcontext(environ=env) as c:
            request = c.request
            request.charset = 'latin1'
            assert request.query['char'] == self.char



class TestEncodeMultipart(object):

    def test_it_encodes_a_data_dict(self):
        data, headers = encode_multipart([('foo', 'bar baf')])
        data = data.getvalue()
        assert b'Content-Disposition: form-data; name="foo"\r\n\r\nbar baf' \
                in data

    def test_it_encodes_a_file_tuple(self):
        data, headers = encode_multipart(
            files=[('foo', 'foo.txt', 'ascii', 'bar')])
        data = data.getvalue()
        expected = (b'Content-Disposition: form-data; '
                    b'name="foo"; filename="foo.txt"\r\n'
                    b'Content-Type: ascii\r\n'
                    b'\r\n'
                    b'bar')
        assert expected in data
